from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import numpy as np
import pytest
import xarray as xr

from glotaran.builtin.io.ascii.wavelength_time_explicit_file import AsciiDataIo
from glotaran.builtin.io.netCDF.netCDF import NetCDFDataIo
from glotaran.builtin.io.sdt.sdt_file_reader import SdtDataIo
from glotaran.io import DataIoInterface
from glotaran.plugin_system.base_registry import PluginOverwriteWarning
from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.data_io_registration import data_io_plugin_table
from glotaran.plugin_system.data_io_registration import get_data_io
from glotaran.plugin_system.data_io_registration import get_dataloader
from glotaran.plugin_system.data_io_registration import get_datasaver
from glotaran.plugin_system.data_io_registration import is_known_data_format
from glotaran.plugin_system.data_io_registration import known_data_formats
from glotaran.plugin_system.data_io_registration import load_dataset
from glotaran.plugin_system.data_io_registration import register_data_io
from glotaran.plugin_system.data_io_registration import save_dataset
from glotaran.plugin_system.data_io_registration import set_data_plugin
from glotaran.plugin_system.data_io_registration import show_data_io_method_help
from glotaran.plugin_system.data_io_registration import supported_file_extensions_data_io
from glotaran.testing.plugin_system import monkeypatch_plugin_registry_data_io

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from _pytest.capture import CaptureFixture

    from glotaran.typing import StrOrPath


class MockDataIo(DataIoInterface):
    def load_dataset(  # type:ignore[override]
        self, file_name: StrOrPath, *, result_container: dict[str, Any], **kwargs: Any
    ) -> xr.Dataset | xr.DataArray:
        """This docstring is just for help testing of 'load_dataset'."""
        result_container |= {"file_name": file_name, **kwargs}
        return xr.DataArray([1, 2])

    # TODO: Investigate why this raises an [override] type error and read_dataset doesn't
    def save_dataset(  # type:ignore[override]
        self,
        file_name: StrOrPath,
        dataset: xr.Dataset | xr.DataArray,
        *,
        result_container: dict[str, Any],
        **kwargs: Any,
    ) -> None:
        result_container.update(
            **{
                "file_name": file_name,
                "dataset": dataset,
                **kwargs,
            }
        )


class MockDataIoPartial(DataIoInterface):
    def load_dataset(self, file_name: StrOrPath) -> xr.Dataset | xr.DataArray:
        return xr.DataArray([1, 2])


MOCK_REGISTRY_VALUES = {
    "foo": DataIoInterface("foo"),
    "mock": MockDataIo("bar"),
    "test_data_io_registration.MockDataIo_bar": MockDataIo("bar"),
}


@pytest.fixture(scope="function")
def mocked_registry():
    with monkeypatch_plugin_registry_data_io(MOCK_REGISTRY_VALUES, create_new_registry=True):
        yield


@pytest.mark.usefixtures("mocked_registry")
def test_register_data_io():
    """Registered data_io plugin is in registry"""

    @register_data_io("dummy")
    class Dummy(DataIoInterface):
        pass

    @register_data_io(["dummy2", "dummy3"])
    class Dummy2(DataIoInterface):
        pass

    for format_name, plugin_class in [("dummy", Dummy), ("dummy2", Dummy2), ("dummy3", Dummy2)]:
        assert format_name in __PluginRegistry.data_io
        assert isinstance(__PluginRegistry.data_io[format_name], plugin_class)
        assert __PluginRegistry.data_io[format_name].format == format_name
        assert isinstance(
            __PluginRegistry.data_io[
                f"test_data_io_registration.{plugin_class.__name__}_{format_name}"
            ],
            plugin_class,
        )


@pytest.mark.usefixtures("mocked_registry")
def test_register_data_io_warning():
    """PluginOverwriteWarning raised pointing to correct file."""

    with pytest.warns(PluginOverwriteWarning, match="Dummy.+dummy.+Dummy2") as record:

        @register_data_io("dummy")
        class Dummy(DataIoInterface):
            pass

        @register_data_io("dummy")
        class Dummy2(DataIoInterface):
            pass

        assert len(record) == 1
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("mocked_registry")
def test_is_known_data_format():
    """Known format in mocked register"""
    assert is_known_data_format("foo")
    assert is_known_data_format("mock")
    assert not is_known_data_format("baz")


def test_known_data_format_actual_register():
    """Builtins in are in actual register."""
    assert is_known_data_format("sdt")
    assert is_known_data_format("ascii")
    assert is_known_data_format("nc")
    assert is_known_data_format("glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo_sdt")
    assert is_known_data_format(
        "glotaran.builtin.io.ascii.wavelength_time_explicit_file.AsciiDataIo_ascii"
    )
    assert is_known_data_format("glotaran.builtin.io.netCDF.netCDF.NetCDFDataIo_nc")


@pytest.mark.parametrize(
    "format_name, io_class",
    (
        ("sdt", SdtDataIo),
        ("ascii", AsciiDataIo),
        ("nc", NetCDFDataIo),
    ),
)
def test_get_data_io(format_name: str, io_class: type[DataIoInterface]):
    """Get the right instance"""
    assert isinstance(get_data_io(format_name), io_class)
    assert get_data_io(format_name).format == format_name


@pytest.mark.usefixtures("mocked_registry")
def test_known_data_formats():
    """Known formats are the same as mocked register keys"""
    assert sorted(known_data_formats()) == sorted(["foo", "mock"])


@pytest.mark.usefixtures("mocked_registry")
def test_set_data_plugin():
    """Set Change Plugin used for format foo"""
    assert isinstance(get_data_io("foo"), DataIoInterface)
    set_data_plugin("foo", "test_data_io_registration.MockDataIo_bar")
    assert isinstance(get_data_io("foo"), MockDataIo)


@pytest.mark.parametrize(
    "format_name, io_class",
    (
        ("sdt", SdtDataIo),
        ("ascii", AsciiDataIo),
        ("nc", NetCDFDataIo),
    ),
)
def test_get_dataloader(format_name: str, io_class: type[DataIoInterface]):
    """Code of the dataloader is the same as original classes method code"""
    dataloader = get_dataloader(format_name)
    assert dataloader.__code__ == io_class.load_dataset.__code__


@pytest.mark.parametrize(
    "format_name, io_class",
    (
        ("sdt", SdtDataIo),
        ("ascii", AsciiDataIo),
        ("nc", NetCDFDataIo),
    ),
)
def test_get_datawriter(format_name: str, io_class: type[DataIoInterface]):
    """Code of the datawriter is the same as original classes method code"""
    datawriter = get_datasaver(format_name)
    assert datawriter.__code__ == io_class.save_dataset.__code__


@pytest.mark.usefixtures("mocked_registry")
def test_load_dataset(tmp_path: Path):
    """All args and kwargs are passes correctly."""
    file_path = tmp_path / "dummy.mock"
    file_path.write_text("mock")
    result: dict[str, Any] = {}

    dataset = load_dataset(file_path, result_container=result, dummy_arg="baz")

    assert result == {"file_name": file_path.as_posix(), "dummy_arg": "baz"}
    assert np.all(dataset.data == xr.DataArray([1, 2]).to_dataset(name="data").data)
    assert dataset.source_path == file_path.as_posix()


@pytest.mark.usefixtures("mocked_registry")
def test_protect_from_overwrite_write_functions(tmp_path: Path):
    """Raise FileExistsError if file exists."""

    file_path = tmp_path / "dummy.foo"
    file_path.touch()

    with pytest.raises(FileExistsError, match="The file .+? already exists"):
        save_dataset(xr.DataArray([1, 2]), str(file_path))


@pytest.mark.parametrize("sub_dir", ("", "sub_dir"))
@pytest.mark.usefixtures("mocked_registry")
def test_write_dataset(tmp_path: Path, sub_dir: str):
    """All args and kwargs are passes correctly."""
    file_path = tmp_path / sub_dir / "dummy.mock"

    result: dict[str, Any] = {}
    save_dataset(
        xr.DataArray([1, 2]),
        file_path,
        result_container=result,
        dummy_arg="baz",
    )

    assert len(result) == 3
    assert result["file_name"] == file_path.as_posix()
    assert result["dummy_arg"] == "baz"
    assert np.all(result["dataset"] == xr.DataArray([1, 2]))


@pytest.mark.usefixtures("mocked_registry")
def test_write_dataset_error(tmp_path: Path):
    """Raise ValueError if method isn't implemented in the DataIo class."""
    file_path = tmp_path / "dummy.foo"

    with pytest.raises(ValueError, match="Cannot save data with format: 'foo'"):
        save_dataset(xr.DataArray([1, 2]), str(file_path), "foo")

    file_path.touch()

    with pytest.raises(ValueError, match="Cannot read data with format: 'foo'"):
        load_dataset(str(file_path))


@pytest.mark.usefixtures("mocked_registry")
def test_show_data_io_method_help(capsys: CaptureFixture):
    """Same help as when called directly."""
    plugin = MockDataIo("foo")
    help(plugin.load_dataset)
    original_help, _ = capsys.readouterr()

    show_data_io_method_help(format_name="mock", method_name="load_dataset")
    result, _ = capsys.readouterr()

    assert "This docstring is just for help testing of 'load_dataset'." in result
    assert result == original_help


@pytest.mark.usefixtures("mocked_registry")
def test_data_io_plugin_table():
    """Plugin foo supports no function and mock supports all"""
    expected = dedent(
        """\
        |  __Format name__  |  __load_dataset__  |  __save_dataset__  |
        |-------------------|--------------------|--------------------|
        |       `foo`       |         /          |         /          |
        |      `mock`       |         *          |         *          |
        """
    )

    assert f"{data_io_plugin_table()}\n" == expected


@pytest.mark.usefixtures("mocked_registry")
def test_data_io_plugin_table_full():
    """Full Table with all extras"""
    expected = dedent(
        """\
        |              __Format name__               |  __load_dataset__  |  __save_dataset__  |               __Plugin name__               |
        |--------------------------------------------|--------------------|--------------------|---------------------------------------------|
        |                   `foo`                    |         /          |         /          | `glotaran.io.interface.DataIoInterface_foo` |
        |                   `mock`                   |         *          |         *          | `test_data_io_registration.MockDataIo_mock` |
        | `test_data_io_registration.MockDataIo_bar` |         *          |         *          | `test_data_io_registration.MockDataIo_bar`  |
        """  # noqa: E501
    )

    assert f"{data_io_plugin_table(plugin_names=True,full_names=True)}\n" == expected


@pytest.mark.parametrize(
    "method_names, expected",
    (
        (
            "load_dataset",
            [".mock", ".mock_partial"],
        ),
        (
            "save_dataset",
            [".mock"],
        ),
        (
            ["load_dataset", "save_dataset"],
            [".mock"],
        ),
    ),
)
def test_supported_file_extensions_data_io(method_names: str | Sequence[str], expected: list[str]):
    """Extension don't list full plugin name and omit extension that don't support all methods."""
    with monkeypatch_plugin_registry_data_io(
        test_data_io={**MOCK_REGISTRY_VALUES, "mock_partial": MockDataIoPartial},
        create_new_registry=True,
    ):
        assert list(supported_file_extensions_data_io(method_names)) == expected



import click

from glotaran.cli.commands import util


@util.signature_analysis
def validate_cmd(parameters_file: str, model_file: str, scheme_file: str):
    """Validates a model file and optionally a parameter file."""

    if scheme_file is not None:
        scheme = util.load_scheme_file(scheme_file, verbose=True)
        click.echo(scheme.validate())
        return

    if model_file is not None:
        model = util.load_model_file(model_file, verbose=True)
        parameters = None
        if parameters_file is not None:
            parameters = util.load_parameter_file(parameters_file, verbose=True)

        click.echo(model.validate(parameters=parameters))
    else:
        click.echo(
            "Neither analysis scheme nor model file specified. "
            "Type 'glotaran validate --help' for more info."
        )



"""Protocol like type definitions."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Protocol
from typing import TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Mapping
    from collections.abc import Sequence

    from glotaran.typing.types import StrOrPath


class FileLoadableProtocol(Protocol):
    """Protocol class that a file loadable class need adherer to."""

    loader: Callable[
        [StrOrPath | Sequence[StrOrPath] | Mapping[str, StrOrPath]],
        FileLoadableProtocol,
    ]
    source_path: StrOrPath | Sequence[StrOrPath] | Mapping[str, StrOrPath]


FileLoadable = TypeVar("FileLoadable", bound=FileLoadableProtocol)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from IPython.core.formatters import format_display_data

from glotaran.utils.ipython import MarkdownStr
from glotaran.utils.ipython import display_file

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    "raw_str, result_str, syntax",
    (
        ("# Model", "# Model", None),
        ("kinetic:\n  - ['1', 1]", "```yaml\nkinetic:\n  - ['1', 1]\n```", "yaml"),
    ),
)
def test_markdown_str_render(raw_str: str, result_str: str, syntax: str):
    """Rendering"""
    result = MarkdownStr(raw_str, syntax=syntax)

    assert str(result) == result_str
    assert result == result_str

    rendered_result = format_display_data(result)[0]

    assert "text/markdown" in rendered_result
    assert rendered_result["text/markdown"] == result_str
    assert rendered_result["text/plain"] == repr(raw_str)


def test_display_file(tmp_path: Path):
    """str and PathLike give the same result"""
    file_content = "kinetic:\n  - ['1', 1]"
    expected = MarkdownStr(file_content, syntax="yaml")
    yml_file = tmp_path / "test.yml"
    yml_file.write_text(file_content)

    assert display_file(yml_file) == expected
    assert display_file(str(yml_file)) == expected

    yaml_file = tmp_path / "test.yaml"
    yaml_file.write_text(file_content)
    assert display_file(yaml_file) == expected

    assert str(display_file(yaml_file, syntax="json")) == f"```json\n{file_content}\n```"
    assert str(display_file(yaml_file, syntax="")) == f"```\n{file_content}\n```"



from __future__ import annotations

from typing import TYPE_CHECKING

import numba as nb
import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.decay.decay_matrix_gaussian_irf import (
    calculate_decay_matrix_gaussian_irf,
)
from glotaran.builtin.megacomplexes.decay.decay_matrix_gaussian_irf import (
    calculate_decay_matrix_gaussian_irf_on_index,
)
from glotaran.builtin.megacomplexes.decay.irf import IrfMultiGaussian
from glotaran.builtin.megacomplexes.decay.irf import IrfSpectralMultiGaussian
from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import get_dataset_model_model_dimension

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


def index_dependent(dataset_model: DatasetModel) -> bool:
    """Determine if a dataset_model is index dependent.

    Parameters
    ----------
    dataset_model : DatasetModel
        A dataset model instance.

    Returns
    -------
    bool
        Returns True if the dataset_model has an IRF that is index dependent (e.g. has dispersion).
    """
    return (
        isinstance(dataset_model.irf, IrfMultiGaussian) and dataset_model.irf.is_index_dependent()
    )


def calculate_matrix(
    megacomplex: Megacomplex,
    dataset_model: DatasetModel,
    global_axis: ArrayLike,
    model_axis: ArrayLike,
    **kwargs,
):
    compartments = megacomplex.get_compartments(dataset_model)
    initial_concentration = megacomplex.get_initial_concentration(dataset_model)
    k_matrix = megacomplex.get_k_matrix()

    rates = k_matrix.rates(compartments, initial_concentration)

    # init the matrix
    matrix_shape = (
        (global_axis.size, model_axis.size, rates.size)
        if index_dependent(dataset_model)
        else (model_axis.size, rates.size)
    )
    matrix = np.zeros(matrix_shape, dtype=np.float64)

    if index_dependent(dataset_model):
        decay_matrix_implementation_index_dependent(
            matrix, rates, global_axis, model_axis, dataset_model
        )
    else:
        decay_matrix_implementation_index_independent(
            matrix, rates, global_axis, model_axis, dataset_model
        )

    if not np.all(np.isfinite(matrix)):
        raise ValueError(
            f"Non-finite concentrations for K-Matrix '{k_matrix.label}':\n"
            f"{k_matrix.matrix_as_markdown(fill_parameters=True)}"
        )

    # apply A matrix
    matrix = matrix @ megacomplex.get_a_matrix(dataset_model)

    # done
    return compartments, matrix


def collect_megacomplexes(dataset_model: DatasetModel, as_global: bool) -> list[Megacomplex]:
    from glotaran.builtin.megacomplexes.decay.decay_megacomplex import DecayMegacomplex
    from glotaran.builtin.megacomplexes.decay.decay_parallel_megacomplex import (
        DecayParallelMegacomplex,
    )
    from glotaran.builtin.megacomplexes.decay.decay_sequential_megacomplex import (
        DecaySequentialMegacomplex,
    )

    return list(
        filter(
            lambda m: isinstance(
                m, (DecayMegacomplex, DecayParallelMegacomplex, DecaySequentialMegacomplex)
            ),
            dataset_model.global_megacomplex if as_global else dataset_model.megacomplex,
        )
    )


def finalize_data(
    dataset_model: DatasetModel,
    dataset: xr.Dataset,
    is_full_model: bool = False,
    as_global: bool = False,
):
    species_dimension = "decay_species" if as_global else "species"
    if species_dimension in dataset.coords:
        # The first decay megacomplescomplex called will finalize the data for all
        # decay megacomplexes.
        return

    decay_megacomplexes = collect_megacomplexes(dataset_model, as_global)
    global_dimension = dataset.attrs["global_dimension"]
    name = "images" if global_dimension == "pixel" else "spectra"

    all_species = []
    for megacomplex in decay_megacomplexes:
        for species in megacomplex.get_compartments(dataset_model):
            if species not in all_species:
                all_species.append(species)
    retrieve_species_associated_data(
        dataset_model,
        dataset,
        all_species,
        species_dimension,
        global_dimension,
        name,
        is_full_model,
        as_global,
    )
    retrieve_initial_concentration(
        dataset_model,
        dataset,
        species_dimension,
    )
    retrieve_irf(dataset_model, dataset, global_dimension)

    if not is_full_model:
        for megacomplex in decay_megacomplexes:
            retrieve_decay_associated_data(
                megacomplex,
                dataset_model,
                dataset,
                global_dimension,
                name,
            )


def decay_matrix_implementation_index_independent(
    matrix: np.ndarray,
    rates: np.ndarray,
    global_axis: np.ndarray,
    model_axis: np.ndarray,
    dataset_model: DatasetModel,
):
    if isinstance(dataset_model.irf, IrfMultiGaussian):
        (
            centers,
            widths,
            irf_scales,
            shift,
            backsweep,
            backsweep_period,
        ) = dataset_model.irf.parameter(None, global_axis)

        calculate_decay_matrix_gaussian_irf_on_index(
            matrix,
            rates,
            model_axis,
            centers - shift,
            widths,
            irf_scales,
            backsweep,
            backsweep_period,
        )
        if dataset_model.irf.normalize:
            matrix /= np.sum(irf_scales)

    else:
        calculate_decay_matrix_no_irf(matrix, rates, model_axis)


def decay_matrix_implementation_index_dependent(
    matrix: np.ndarray,
    rates: np.ndarray,
    global_axis: np.ndarray,
    model_axis: np.ndarray,
    dataset_model: DatasetModel,
):
    all_centers, all_widths = [], []
    backsweep, backsweep_period = False, None
    irf_scales = []
    for global_index in range(global_axis.size):
        (
            centers,
            widths,
            irf_scales,
            shift,
            backsweep,
            backsweep_period,
        ) = dataset_model.irf.parameter(global_index, global_axis)
        all_centers.append(centers - shift)
        all_widths.append(widths)

    calculate_decay_matrix_gaussian_irf(
        matrix,
        rates,
        model_axis,
        np.array(all_centers),
        np.array(all_widths),
        irf_scales,
        backsweep,
        backsweep_period,
    )
    if dataset_model.irf.normalize:
        matrix /= np.sum(irf_scales)


@nb.jit(nopython=True, parallel=True)
def calculate_decay_matrix_no_irf(matrix, rates, times):
    for n_r in nb.prange(rates.size):
        r_n = rates[n_r]
        for n_t in range(times.size):
            t_n = times[n_t]
            matrix[n_t, n_r] += np.exp(-r_n * t_n)


def retrieve_species_associated_data(
    dataset_model: DatasetModel,
    dataset: xr.Dataset,
    species: list[str],
    species_dimension: str,
    global_dimension: str,
    name: str,
    is_full_model: bool,
    as_global: bool,
):
    model_dimension = get_dataset_model_model_dimension(dataset_model)
    if as_global:
        model_dimension, global_dimension = global_dimension, model_dimension
    dataset.coords[species_dimension] = species

    matrix = dataset.global_matrix if as_global else dataset.matrix
    clp_dim = "global_clp_label" if as_global else "clp_label"

    if len(dataset.matrix.shape) == 3:
        #  index dependent
        dataset["species_concentration"] = (
            (
                global_dimension,
                model_dimension,
                species_dimension,
            ),
            matrix.sel({clp_dim: species}).values,
        )
    else:
        #  index independent
        dataset["species_concentration"] = (
            (
                model_dimension,
                species_dimension,
            ),
            matrix.sel({clp_dim: species}).values,
        )

    if not is_full_model:
        dataset[f"species_associated_{name}"] = (
            (
                global_dimension,
                species_dimension,
            ),
            dataset.clp.sel(clp_label=species).data,
        )


def retrieve_initial_concentration(
    dataset_model: DatasetModel,
    dataset: xr.Dataset,
    species_dimension: str,
):
    if (
        not hasattr(dataset_model, "initial_concentration")
        or dataset_model.initial_concentration is None
    ):
        # For parallel and sequential decay we don't have dataset wide initial concentration
        # unless mixed with general decays
        return

    dataset["initial_concentration"] = (
        (species_dimension,),
        dataset_model.initial_concentration.parameters,
    )


def retrieve_decay_associated_data(
    megacomplex: Megacomplex,
    dataset_model: DatasetModel,
    dataset: xr.Dataset,
    global_dimension: str,
    name: str,
):
    species = megacomplex.get_compartments(dataset_model)
    initial_concentration = megacomplex.get_initial_concentration(dataset_model)
    k_matrix = megacomplex.get_k_matrix()

    matrix = k_matrix.full(species)
    matrix_reduced = k_matrix.reduced(species)
    a_matrix = megacomplex.get_a_matrix(dataset_model)
    rates = k_matrix.rates(species, initial_concentration)
    lifetimes = 1 / rates

    das = dataset[f"species_associated_{name}"].sel(species=species).values @ a_matrix.T

    component_name = f"component_{megacomplex.label}"
    component_coords = {
        component_name: np.arange(1, rates.size + 1),
        f"rate_{megacomplex.label}": (component_name, rates),
        f"lifetime_{megacomplex.label}": (component_name, lifetimes),
    }
    das_coords = component_coords.copy()
    das_coords[global_dimension] = dataset.coords[global_dimension]
    das_name = f"decay_associated_{name}_{megacomplex.label}"
    das = xr.DataArray(das, dims=(global_dimension, component_name), coords=das_coords)

    initial_concentration = megacomplex.get_initial_concentration(dataset_model, normalized=False)
    species_name = f"species_{megacomplex.label}"
    a_matrix_coords = component_coords.copy()
    a_matrix_coords[species_name] = species
    a_matrix_coords[f"initial_concentration_{megacomplex.label}"] = (
        species_name,
        initial_concentration,
    )
    a_matrix_name = f"a_matrix_{megacomplex.label}"
    a_matrix = xr.DataArray(a_matrix, coords=a_matrix_coords, dims=(component_name, species_name))

    to_species_name = f"to_species_{megacomplex.label}"
    from_species_name = f"from_species_{megacomplex.label}"
    k_matrix_name = f"k_matrix_{megacomplex.label}"
    k_matrix = xr.DataArray(
        matrix, coords=[(to_species_name, species), (from_species_name, species)]
    )

    k_matrix_reduced_name = f"k_matrix_reduced_{megacomplex.label}"
    k_matrix_reduced = xr.DataArray(
        matrix_reduced, coords=[(to_species_name, species), (from_species_name, species)]
    )

    dataset[das_name] = das
    dataset[a_matrix_name] = a_matrix
    dataset[k_matrix_name] = k_matrix
    dataset[k_matrix_reduced_name] = k_matrix_reduced


def retrieve_irf(dataset_model: DatasetModel, dataset: xr.Dataset, global_dimension: str):
    if not isinstance(dataset_model.irf, IrfMultiGaussian) or "irf" in dataset:
        return

    irf = dataset_model.irf
    model_dimension = get_dataset_model_model_dimension(dataset_model)

    dataset["irf"] = (
        (model_dimension),
        irf.calculate(
            index=0,
            global_axis=dataset.coords[global_dimension].values,
            model_axis=dataset.coords[model_dimension].values,
        ).data,
    )

    center = irf.center if isinstance(irf.center, list) else [irf.center]
    width = irf.width if isinstance(irf.width, list) else [irf.width]
    dataset["irf_center"] = ("irf_nr", center) if len(center) > 1 else center[0]
    dataset["irf_width"] = ("irf_nr", width) if len(width) > 1 else width[0]

    if irf.shift is not None:
        dataset["irf_shift"] = (global_dimension, [center[0] - p.value for p in irf.shift])

    if isinstance(irf, IrfSpectralMultiGaussian) and irf.dispersion_center:
        dataset["irf_center_location"] = (
            ("irf_nr", global_dimension),
            irf.calculate_dispersion(dataset.coords["spectral"].values),
        )
        # center_dispersion_1 for backwards compatibility (0.3-0.4.1)
        dataset["center_dispersion_1"] = dataset["irf_center_location"].sel(irf_nr=0)



from __future__ import annotations

import ctypes
from typing import TYPE_CHECKING

import numba as nb
import numpy as np
from numba.extending import get_cython_function_address

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike
# This is a work around to use scipy.special function with numba
_dble = ctypes.c_double

functype = ctypes.CFUNCTYPE(_dble, _dble)

erf_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erf")
erfcx_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erfcx")

erf = functype(erf_addr)
erfcx = functype(erfcx_addr)

SQRT2 = np.sqrt(2)


@nb.jit(nopython=True, parallel=False)
def calculate_decay_matrix_gaussian_irf_on_index(
    matrix: ArrayLike,
    rates: ArrayLike,
    times: ArrayLike,
    centers: ArrayLike,
    widths: ArrayLike,
    scales: ArrayLike,
    backsweep: bool,
    backsweep_period: float | None,
):
    """Calculates a decay matrix with a gaussian irf."""
    for n_i in nb.prange(centers.size):
        center, width, scale = centers[n_i], widths[n_i], scales[n_i]
        for n_r in nb.prange(rates.size):
            r_n = rates[n_r]
            backsweep_valid = backsweep and abs(r_n) * backsweep_period > 0.001
            alpha = (r_n * width) / SQRT2
            for n_t in nb.prange(times.size):
                t_n = times[n_t]
                beta = (t_n - center) / (width * SQRT2)
                thresh = beta - alpha
                if thresh < -1:
                    matrix[n_t, n_r] += scale * 0.5 * erfcx(-thresh) * np.exp(-beta * beta)
                else:
                    matrix[n_t, n_r] += (
                        scale * 0.5 * (1 + erf(thresh)) * np.exp(alpha * (alpha - 2 * beta))
                    )
                if backsweep and backsweep_valid:
                    x1 = np.exp(-r_n * (t_n - center + backsweep_period))
                    x2 = np.exp(-r_n * ((backsweep_period / 2) - (t_n - center)))
                    x3 = np.exp(-r_n * backsweep_period)
                    matrix[n_t, n_r] += scale * (x1 + x2) / (1 - x3)


@nb.jit(nopython=True, parallel=True)
def calculate_decay_matrix_gaussian_irf(
    matrix: ArrayLike,
    rates: ArrayLike,
    times: ArrayLike,
    all_centers: ArrayLike,
    all_widths: ArrayLike,
    scales: ArrayLike,
    backsweep: bool,
    backsweep_period: float | None,
):
    for n_w in nb.prange(all_centers.shape[0]):
        calculate_decay_matrix_gaussian_irf_on_index(
            matrix[n_w],
            rates,
            times,
            all_centers[n_w],
            all_widths[n_w],
            scales,
            backsweep,
            backsweep_period,
        )



"""This package contains the decay megacomplex item."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.decay.irf import Irf
from glotaran.builtin.megacomplexes.decay.k_matrix import KMatrix
from glotaran.builtin.megacomplexes.decay.util import calculate_matrix
from glotaran.builtin.megacomplexes.decay.util import finalize_data
from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import ModelItemType
from glotaran.model import ParameterType
from glotaran.model import item
from glotaran.model import megacomplex

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@item
class DecayDatasetModel(DatasetModel):
    irf: ModelItemType[Irf] | None = None


@megacomplex(dataset_model_type=DecayDatasetModel)
class DecayParallelMegacomplex(Megacomplex):
    dimension: str = "time"
    type: str = "decay-parallel"
    compartments: list[str]
    rates: list[ParameterType]

    def get_compartments(self, dataset_model: DatasetModel) -> list[str]:
        return self.compartments

    def get_initial_concentration(
        self, dataset_model: DatasetModel, normalized: bool = True
    ) -> np.ndarray:
        initial_concentration = np.ones((len(self.compartments)), dtype=np.float64)
        if normalized:
            initial_concentration /= initial_concentration.size
        return initial_concentration

    def get_k_matrix(self) -> KMatrix:
        return KMatrix(
            label="",
            matrix={
                (self.compartments[i], self.compartments[i]): self.rates[i]
                for i in range(len(self.compartments))
            },
        )

    def get_a_matrix(self, dataset_model: DatasetModel) -> np.ndarray:
        return self.get_k_matrix().a_matrix_general(
            self.get_compartments(dataset_model), self.get_initial_concentration(dataset_model)
        )

    def calculate_matrix(
        self,
        dataset_model: DecayDatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        return calculate_matrix(self, dataset_model, global_axis, model_axis, **kwargs)

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        finalize_data(dataset_model, dataset, is_full_model, as_global)



"""Module containing the data provider classes."""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.model import DatasetGroup
from glotaran.model import Model
from glotaran.model.dataset_model import get_dataset_model_model_dimension
from glotaran.model.dataset_model import has_dataset_model_global_model
from glotaran.project import Scheme

if TYPE_CHECKING:
    from typing import Literal

    from glotaran.typing.types import ArrayLike


class AlignDatasetError(ValueError):
    """Indicates that datasets can not be aligned."""

    def __init__(self):
        """Initialize a AlignDatasetError."""
        super().__init__(
            "Cannot link datasets, aligning is ambiguous. \n\n"
            "Try to lower link tolerance or change the alignment method."
        )


class DataProvider:
    """A class to provide prepared data for optimization."""

    def __init__(self, scheme: Scheme, dataset_group: DatasetGroup):
        """Initialize a data provider for a scheme and a dataset_group.

        Parameters
        ----------
        scheme : Scheme
            The optimization scheme.
        dataset_group : DatasetGroup
            The dataset group.
        """
        self._data: dict[str, ArrayLike] = {}
        self._weight: dict[str, ArrayLike | None] = {}
        self._flattened_data: dict[str, ArrayLike] = {}
        self._flattened_weight: dict[str, ArrayLike | None] = {}
        self._model_axes: dict[str, ArrayLike] = {}
        self._model_dimensions: dict[str, str] = {}
        self._global_axes: dict[str, ArrayLike] = {}
        self._global_dimensions: dict[str, str] = {}

        for label, dataset_model in dataset_group.dataset_models.items():
            dataset = scheme.data[label]
            model_dimension = get_dataset_model_model_dimension(dataset_model)
            self._model_axes[label] = dataset.coords[model_dimension].data
            self._model_dimensions[label] = model_dimension
            global_dimension = self.infer_global_dimension(model_dimension, dataset.data.dims)
            self._global_axes[label] = dataset.coords[global_dimension].data
            self._global_dimensions[label] = global_dimension

            self._weight[label] = self.get_from_dataset(
                dataset, "weight", model_dimension, global_dimension
            )
            self.add_model_weight(scheme.model, label, model_dimension, global_dimension)

            self._data[label] = self.get_from_dataset(  # type:ignore[assignment]
                dataset, "data", model_dimension, global_dimension
            )
            if self._weight[label] is not None:
                self._data[label] *= self._weight[label]

            if has_dataset_model_global_model(dataset_model):
                self._flattened_data[label] = self._data[label].T.flatten()
                self._flattened_weight[label] = (
                    self._weight[label].T.flatten()  # type:ignore[union-attr]
                    if self._weight[label] is not None
                    else None
                )

    @staticmethod
    def infer_global_dimension(model_dimension: str, dimensions: tuple[str]) -> str:
        """Infer the name of the global dimension from tuple of dimensions.

        Parameters
        ----------
        model_dimension : str
            The model dimension.
        dimensions : tuple[str]
            The dimensions tuple to infer from.

        Returns
        -------
        str
            The inferred name of the global dimension.
        """
        return next(dim for dim in dimensions if dim != model_dimension)

    @staticmethod
    def get_from_dataset(
        dataset: xr.Dataset, name: str, model_dimension: str, global_dimension: str
    ) -> ArrayLike | None:
        """Get a copy of data from a dataset with dimensions (model, global).

        Parameters
        ----------
        dataset : xr.Dataset
            The dataset to retrieve from.
        name : str
            The name of the data to retrieve.
        model_dimension : str
            The model dimension.
        global_dimension : str
            The global dimension.

        Returns
        -------
        ArrayLike | None
            The copy of the data. None if name is not present in dataset.
        """
        data = None
        if name in dataset:
            data = dataset[name].data.copy()
            if dataset[name].dims != (model_dimension, global_dimension):
                data = data.T
        return data

    @staticmethod
    def get_axis_slice_from_interval(interval: tuple[float, float], axis: ArrayLike) -> slice:
        """Get a slice of indices from a min max tuple and for an axis.

        Parameters
        ----------
        interval : tuple[float, float]
            The min max tuple.
        axis : ArrayLike
            The axis to slice.

        Returns
        -------
        slice
            The slice of indices.
        """
        interval_min = interval[0]
        interval_max = interval[1]

        if interval_min > interval_max:
            interval_min, interval_max = interval_max, interval_min

        minimum = 0 if np.isinf(interval_min) else np.abs(axis - interval_min).argmin()
        maximum = (
            axis.size - 1 if np.isinf(interval_max) else np.abs(axis - interval_max).argmin() + 1
        )

        return slice(minimum, maximum)

    def add_model_weight(
        self,
        model: Model,
        dataset_label: str,
        model_dimension: str,
        global_dimension: str,
    ):
        """Add model weight to data.

        Parameters
        ----------
        model : Model
            The model.
        dataset_label : str
            The label of the data.
        model_dimension : str
            The model dimension.
        global_dimension : str
            The global dimension.
        """
        model_weights = [weight for weight in model.weights if dataset_label in weight.datasets]
        if not model_weights:
            return

        if self._weight[dataset_label]:
            warnings.warn(
                f"Ignoring model weight for dataset '{dataset_label}'"
                " because weight is already supplied by dataset."
            )
            return
        model_axis = self._model_axes[dataset_label]
        global_axis = self._global_axes[dataset_label]
        weight = xr.DataArray(
            np.ones((model_axis.size, global_axis.size)),
            coords=(
                (model_dimension, model_axis),
                (global_dimension, global_axis),
            ),
        )
        for model_weight in model_weights:
            idx = {}
            if model_weight.global_interval is not None:
                idx[global_dimension] = self.get_axis_slice_from_interval(
                    model_weight.global_interval, global_axis
                )
            if model_weight.model_interval is not None:
                idx[model_dimension] = self.get_axis_slice_from_interval(
                    model_weight.model_interval, model_axis
                )
            weight[idx] *= model_weight.value

        self._weight[dataset_label] = weight.data

    def get_data(self, dataset_label: str) -> ArrayLike:
        """Get data for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        ArrayLike
            The data.
        """
        return self._data[dataset_label]

    def get_weight(self, dataset_label: str) -> ArrayLike | None:
        """Get weight for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        ArrayLike | None
            The weight.
        """
        return self._weight[dataset_label]

    def get_flattened_data(self, dataset_label: str) -> ArrayLike:
        """Get flattened data for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        ArrayLike
            The flattened data.
        """
        return self._flattened_data[dataset_label]

    def get_flattened_weight(self, dataset_label: str) -> ArrayLike | None:
        """Get flattened weight for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        ArrayLike | None
            The flattened weight.
        """
        return self._flattened_weight[dataset_label]

    def get_model_axis(self, dataset_label: str) -> ArrayLike:
        """Get the model axis for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        ArrayLike
            The model axis.
        """
        return self._model_axes[dataset_label]

    def get_model_dimension(self, dataset_label: str) -> str:
        """Get the model dimension for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        str
            The model dimension.
        """
        return self._model_dimensions[dataset_label]

    def get_global_axis(self, dataset_label: str) -> ArrayLike:
        """Get the global axis for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        ArrayLike
            The global axis.
        """
        return self._global_axes[dataset_label]

    def get_global_dimension(self, dataset_label: str) -> str:
        """Get the global dimension for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the data.

        Returns
        -------
        str
            The global dimension.
        """
        return self._global_dimensions[dataset_label]


class DataProviderLinked(DataProvider):
    """A class to provide aligned data for optimization."""

    def __init__(
        self,
        scheme: Scheme,
        dataset_group: DatasetGroup,
    ):
        """Initialize a linked data provider for a scheme and a dataset_group.

        Parameters
        ----------
        scheme : Scheme
            The optimization scheme.
        dataset_group : DatasetGroup
            The dataset group.
        """
        super().__init__(scheme, dataset_group)
        aligned_global_axes = self.create_aligned_global_axes(scheme)
        self._aligned_global_axis, self._aligned_data = self.align_data(aligned_global_axes)
        self._aligned_dataset_indices = self.align_dataset_indices(aligned_global_axes)
        self._aligned_group_labels, self._group_definitions = self.align_groups(
            aligned_global_axes
        )
        self._aligned_weights = self.align_weights(aligned_global_axes)

    @staticmethod
    def align_index(
        index: int,
        target_axis: ArrayLike,
        tolerance: float,
        method: Literal["nearest", "backward", "forward"],
    ) -> int:
        """Align an index on a target axis.

        Parameters
        ----------
        index : int
            The index to align.
        target_axis : ArrayLike
            The axis to align the index on.
        tolerance : float
            The alignment tolerance.
        method : Literal["nearest", "backward", "forward"]
            The alignment method.

        Returns
        -------
        int
            The aligned index.
        """
        diff = target_axis - index

        if method == "forward":
            diff = diff[diff >= 0]
        elif method == "backward":
            diff = diff[diff <= 0]

        diff = np.abs(diff)

        if len(diff) > 0 and diff.min() <= tolerance:
            index = target_axis[diff.argmin()]
        return index

    @property
    def aligned_global_axis(self) -> ArrayLike:
        """Get the aligned global axis for the dataset group.

        Returns
        -------
        ArrayLike
            The aligned global axis.
        """
        return self._aligned_global_axis

    @property
    def group_definitions(self) -> dict[str, list[str]]:
        """Get the group definitions for the dataset group.

        Returns
        -------
        dict[str, list[str]]
            The group definitions.
        """
        return self._group_definitions

    def get_aligned_group_label(self, index: int) -> str:
        """Get the group label for an index.

        Parameters
        ----------
        index : int
            The index on the aligned global axis.

        Returns
        -------
        str
            The aligned group label.
        """
        return self._aligned_group_labels[index]

    def get_aligned_dataset_indices(self, index: int) -> ArrayLike:
        """Get the aligned dataset indices for an index.

        Parameters
        ----------
        index : int
            The index on the aligned global axis.

        Returns
        -------
        ArrayLike
            The aligned dataset indices.
        """
        return self._aligned_dataset_indices[index]

    def get_aligned_data(self, index: int) -> ArrayLike:
        """Get the aligned data for an index.

        Parameters
        ----------
        index : int
            The index on the aligned global axis.

        Returns
        -------
        ArrayLike
            The aligned data.
        """
        return self._aligned_data[index]

    def get_aligned_weight(self, index: int) -> ArrayLike | None:
        """Get the aligned weight for an index.

        Parameters
        ----------
        index : int
            The index on the aligned global axis.

        Returns
        -------
        ArrayLike | None
            The aligned weight.
        """
        return self._aligned_weights[index]

    def create_aligned_global_axes(self, scheme: Scheme) -> dict[str, ArrayLike]:
        """Create aligned global axes for the dataset group.

        Parameters
        ----------
        scheme : Scheme
            The optimization scheme.

        Returns
        -------
        dict[str, ArrayLike]
            The aligned global axes.

        Raises
        ------
        AlignDatasetError
            Raised when dataset alignment is ambiguous.
        """
        aligned_axis_values = None
        aligned_global_axes = {}
        for label, global_axis in self._global_axes.items():
            aligned_global_axis = global_axis
            if aligned_axis_values is None:
                aligned_axis_values = aligned_global_axis
            else:
                aligned_global_axis = [
                    self.align_index(
                        index,
                        aligned_axis_values,
                        scheme.clp_link_tolerance,
                        scheme.clp_link_method,
                    )
                    for index in aligned_global_axis
                ]
                if len(np.unique(aligned_global_axis)) != len(aligned_global_axis):
                    raise AlignDatasetError()
                aligned_axis_values = np.unique(
                    np.concatenate([aligned_axis_values, aligned_global_axis])
                )
            aligned_global_axes[label] = aligned_global_axis
        return aligned_global_axes

    def align_data(
        self, aligned_global_axes: dict[str, ArrayLike]
    ) -> tuple[ArrayLike, list[ArrayLike]]:
        """Align the data in a dataset group.

        Parameters
        ----------
        aligned_global_axes : dict[str, ArrayLike]
            The aligned global axes.

        Returns
        -------
        tuple[ArrayLike, list[ArrayLike]]
            The aligned global axis and data.
        """
        aligned_data = xr.concat(
            [
                xr.DataArray(
                    self.get_data(label),
                    dims=["model", "global"],
                    coords={"global": axis},
                )
                for label, axis in aligned_global_axes.items()
            ],
            dim="model",
        )
        aligned_global_axis = aligned_data.coords["global"].data
        return (
            aligned_global_axis,
            [
                aligned_data.isel({"global": i}).dropna(dim="model").data
                for i in range(aligned_global_axis.size)
            ],
        )

    def align_dataset_indices(self, aligned_global_axes: dict[str, ArrayLike]) -> list[ArrayLike]:
        """Align the global indices in a dataset group.

        Parameters
        ----------
        aligned_global_axes : dict[str, ArrayLike]
            The aligned global axes.

        Returns
        -------
        list[ArrayLike]
        The aligned dataset indices.
        """
        aligned_indices = xr.concat(
            [
                xr.DataArray(
                    np.arange(len(axis), dtype=int),
                    dims=["global"],
                    coords={"global": axis},
                )
                for axis in aligned_global_axes.values()
            ],
            dim="dataset",
        )
        return [
            aligned_indices.isel({"global": i}).dropna(dim="dataset").data.astype(int)
            for i in range(self._aligned_global_axis.size)
        ]

    @staticmethod
    def align_groups(
        aligned_global_axes: dict[str, ArrayLike]
    ) -> tuple[ArrayLike, dict[str, list[str]]]:
        """Align the groups in a dataset group.

        Parameters
        ----------
        aligned_global_axes : dict[str, ArrayLike]
            The aligned global axes.

        Returns
        -------
        tuple[ArrayLike, dict[str, list[str]]]
            The aligned grouplabels and group definitions.
        """
        aligned_groups = xr.concat(
            [
                xr.DataArray(np.full(len(axis), label), dims=["global"], coords={"global": axis})
                for label, axis in aligned_global_axes.items()
            ],
            dim="dataset",
            fill_value="",
        )
        # for every element along the global axis, concatenate all dataset labels
        # into an ndarray of shape (len(global,)
        # as an alternative to the more elegant xarray built-in which is limited to 32 datasets
        # aligned_group_labels = aligned_groups.str.join(dim="dataset").data
        aligned_group_labels = np.asarray(
            ["".join(sub_arr.values) for _, sub_arr in aligned_groups.groupby("global")]
        )

        group_definitions: dict[str, list[str]] = {}
        for i, group_label in enumerate(aligned_group_labels):
            if group_label not in group_definitions:
                group_definitions[group_label] = list(
                    filter(lambda label: label != "", aligned_groups.isel({"global": i}).data)
                )
        return aligned_group_labels, group_definitions

    def align_weights(self, aligned_global_axes: dict[str, ArrayLike]) -> list[ArrayLike | None]:
        """Align the weights in a dataset group.

        Parameters
        ----------
        aligned_global_axes : dict[str, ArrayLike]
            The aligned global axes.

        Returns
        -------
        list[ArrayLike | None]
            The aligned weights.
        """
        all_weights = {
            label: xr.DataArray(
                weight,
                dims=["model", "global"],
                coords={"global": aligned_global_axes[label]},
            )
            for label, weight in self._weight.items()
            if weight is not None
        }

        aligned_weights: list[ArrayLike | None] = [None] * self._aligned_global_axis.size
        if all_weights:
            for i, group_label in enumerate(self._aligned_group_labels):
                group_dataset_labels = self._group_definitions[group_label]
                if any(label in all_weights for label in group_dataset_labels):
                    index_weights = []
                    for label in group_dataset_labels:
                        if label in all_weights:
                            index_weights.append(
                                all_weights[label]
                                .sel({"global": self._aligned_global_axis[i]})
                                .data
                            )
                        else:
                            size = self.get_model_axis(label).size
                            index_weights.append(np.ones(size))
                    aligned_weights[i] = np.concatenate(index_weights)

        return aligned_weights



"""Tests for ``glotaran.optimization.optimizer``."""

from textwrap import dedent

import pytest

from glotaran.optimization.optimizer import Optimizer


@pytest.mark.parametrize(
    "optimize_stdout, expected",
    (
        ("random string", 0),
        (
            dedent(
                """\
                   Iteration     Total nfev        Cost      Cost reduction    Step norm     Optimality
                """  # noqa: E501
            ),
            0,
        ),
        (
            dedent(
                """\
                   Iteration     Total nfev        Cost      Cost reduction    Step norm     Optimality
                       0              1         7.5834e+00                                    3.84e+01
                       1              2         7.5833e+00      1.37e-04       4.55e-05       1.26e-01
                """  # noqa: E501
            ),
            1,
        ),
        (
            dedent(
                """\
                   Iteration     Total nfev        Cost      Cost reduction    Step norm     Optimality
                       0              1         7.5834e+00                                    3.84e+01
                       1              2         7.5833e+00      1.37e-04       4.55e-05       1.26e-01
                       2              3         7.5833e+00      6.02e-11       6.44e-09       1.64e-05
                Both `ftol` and `xtol` termination conditions are satisfied.
                Function evaluations 3, initial cost 7.5834e+00, final cost 7.5833e+00, first-order optimality 1.64e-05.
                """  # noqa: E501
            ),
            2,
        ),
    ),
)
def test_optimizer_get_current_optimization_iteration(optimize_stdout: str, expected: int):
    """Test that the correct iteration is returned."""
    assert Optimizer.get_current_optimization_iteration(optimize_stdout) == expected



"""The glotaran generator module."""
from __future__ import annotations

from collections.abc import Callable
from typing import Any
from typing import TypedDict
from typing import cast

from glotaran.builtin.io.yml.utils import write_dict
from glotaran.builtin.megacomplexes.decay import DecayParallelMegacomplex
from glotaran.builtin.megacomplexes.decay import DecaySequentialMegacomplex
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.model import Model


def _generate_decay_model(
    *, nr_compartments: int, irf: bool, spectral: bool, decay_type: str
) -> dict[str, Any]:
    """Generate a decay model dictionary.

    Parameters
    ----------
    nr_compartments : int
        The number of compartments.
    irf : bool
        Whether to add a gaussian irf.
    spectral : bool
        Whether to add a spectral model.
    decay_type : str
        The dype of the decay

    Returns
    -------
    dict[str, Any] :
        The generated model dictionary.
    """
    compartments = [f"species_{i+1}" for i in range(nr_compartments)]
    rates = [f"rates.species_{i+1}" for i in range(nr_compartments)]
    model = {
        "megacomplex": {
            f"megacomplex_{decay_type}_decay": {
                "type": f"decay-{decay_type}",
                "compartments": compartments,
                "rates": rates,
            },
        },
        "dataset": {"dataset_1": {"megacomplex": [f"megacomplex_{decay_type}_decay"]}},
    }
    if spectral:
        model["megacomplex"]["megacomplex_spectral"] = {  # type:ignore[index]
            "type": "spectral",
            "shape": {
                compartment: f"shape_species_{i+1}" for i, compartment in enumerate(compartments)
            },
        }
        model["shape"] = {
            f"shape_species_{i+1}": {
                "type": "gaussian",
                "amplitude": f"shapes.species_{i+1}.amplitude",
                "location": f"shapes.species_{i+1}.location",
                "width": f"shapes.species_{i+1}.width",
            }
            for i in range(nr_compartments)
        }
        model["dataset"]["dataset_1"]["global_megacomplex"] = [  # type:ignore[index]
            "megacomplex_spectral"
        ]
    if irf:
        model["dataset"]["dataset_1"]["irf"] = "gaussian_irf"  # type:ignore[index]
        model["irf"] = {
            "gaussian_irf": {"type": "gaussian", "center": "irf.center", "width": "irf.width"},
        }
    return model


def generate_parallel_decay_model(
    *, nr_compartments: int = 1, irf: bool = False
) -> dict[str, Any]:
    """Generate a parallel decay model dictionary.

    Parameters
    ----------
    nr_compartments : int
        The number of compartments.
    irf : bool
        Whether to add a gaussian irf.

    Returns
    -------
    dict[str, Any] :
        The generated model dictionary.
    """
    return _generate_decay_model(
        nr_compartments=nr_compartments, irf=irf, spectral=False, decay_type="parallel"
    )


def generate_parallel_spectral_decay_model(
    *, nr_compartments: int = 1, irf: bool = False
) -> dict[str, Any]:
    """Generate a parallel spectral decay model dictionary.

    Parameters
    ----------
    nr_compartments : int
        The number of compartments.
    irf : bool
        Whether to add a gaussian irf.

    Returns
    -------
    dict[str, Any] :
        The generated model dictionary.
    """
    return _generate_decay_model(
        nr_compartments=nr_compartments, irf=irf, spectral=True, decay_type="parallel"
    )


def generate_sequential_decay_model(nr_compartments: int = 1, irf: bool = False) -> dict[str, Any]:
    """Generate a sequential decay model dictionary.

    Parameters
    ----------
    nr_compartments : int
        The number of compartments.
    irf : bool
        Whether to add a gaussian irf.

    Returns
    -------
    dict[str, Any] :
        The generated model dictionary.
    """
    return _generate_decay_model(
        nr_compartments=nr_compartments, irf=irf, spectral=False, decay_type="sequential"
    )


def generate_sequential_spectral_decay_model(
    *, nr_compartments: int = 1, irf: bool = False
) -> dict[str, Any]:
    """Generate a sequential spectral decay model dictionary.

    Parameters
    ----------
    nr_compartments : int
        The number of compartments.
    irf : bool
        Whether to add a gaussian irf.

    Returns
    -------
    dict[str, Any] :
        The generated model dictionary.
    """
    return _generate_decay_model(
        nr_compartments=nr_compartments, irf=irf, spectral=True, decay_type="sequential"
    )


generators: dict[str, Callable] = {
    "decay_parallel": generate_parallel_decay_model,
    "spectral_decay_parallel": generate_parallel_spectral_decay_model,
    "decay_sequential": generate_sequential_decay_model,
    "spectral_decay_sequential": generate_sequential_spectral_decay_model,
}

available_generators: list[str] = list(generators.keys())


class GeneratorArguments(TypedDict, total=False):
    """Arguments used by ``generate_model`` and ``generate_model``.

    Parameters
    ----------
    nr_compartments : int
        The number of compartments.
    irf : bool
        Whether to add a gaussian irf.

    See Also
    --------
    generate_model
    generate_model_yml
    """

    nr_compartments: int
    irf: bool


def generate_model(*, generator_name: str, generator_arguments: GeneratorArguments) -> Model:
    """Generate a model.

    Parameters
    ----------
    generator_name : str
        The generator to use.
    generator_arguments : GeneratorArguments
        Arguments for the generator.

    Returns
    -------
    Model
        The generated model

    See Also
    --------
    generate_parallel_decay_model
    generate_parallel_spectral_decay_model
    generate_sequential_decay_model
    generate_sequential_spectral_decay_model

    Raises
    ------
    ValueError
        Raised when an unknown generator is specified.
    """
    if generator_name not in generators:
        raise ValueError(
            f"Unknown model generator '{generator_name}'. "
            f"Known generators are: {list(generators.keys())}"
        )
    model = generators[generator_name](**generator_arguments)
    return Model.create_class_from_megacomplexes(
        [DecayParallelMegacomplex, DecaySequentialMegacomplex, SpectralMegacomplex]
    )(**model)


def generate_model_yml(*, generator_name: str, generator_arguments: GeneratorArguments) -> str:
    """Generate a model as yml string.

    Parameters
    ----------
    generator_name : str
        The generator to use.
    generator_arguments : GeneratorArguments
        Arguments for the generator.

    Returns
    -------
    str
        The generated model yml string.

    See Also
    --------
    generate_parallel_decay_model
    generate_parallel_spectral_decay_model
    generate_sequential_decay_model
    generate_sequential_spectral_decay_model

    Raises
    ------
    ValueError
        Raised when an unknown generator is specified.
    """
    if generator_name not in generators:
        raise ValueError(
            f"Unknown model generator '{generator_name}'. "
            f"Known generators are: {list(generators.keys())}"
        )
    model = generators[generator_name](**generator_arguments)
    return cast(str, write_dict(model))



import numpy as np

from glotaran.builtin.megacomplexes.baseline import BaselineMegacomplex
from glotaran.model import Model
from glotaran.model import fill_item
from glotaran.optimization.matrix_provider import MatrixProvider
from glotaran.parameter import Parameters


def test_baseline():
    model = Model.create_class_from_megacomplexes([BaselineMegacomplex])(
        **{
            "megacomplex": {"m": {"type": "baseline", "dimension": "time"}},
            "dataset": {"dataset1": {"megacomplex": ["m"]}},
        }
    )

    parameters = Parameters({})
    time = np.asarray(np.arange(0, 50, 1.5))
    pixel = np.asarray([0])
    dataset_model = fill_item(model.dataset["dataset1"], model, parameters)
    matrix = MatrixProvider.calculate_dataset_matrix(dataset_model, pixel, time)
    compartments = matrix.clp_labels

    assert len(compartments) == 1
    assert "dataset1_baseline" in compartments

    assert matrix.matrix.shape == (time.size, 1)
    assert np.all(matrix.matrix[:, 0] == 1)



"""Functionality to register, initialize and retrieve glotaran plugins.

Since this module is imported at the root ``__init__.py`` file all other
glotaran imports should be used for typechecking only in the 'if TYPE_CHECKING' block.
This is to prevent issues with circular imports.
"""
from __future__ import annotations

import os
from collections.abc import Iterable
from importlib import metadata
from typing import TYPE_CHECKING
from typing import cast
from warnings import warn

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Generator
    from collections.abc import MutableMapping
    from collections.abc import Sequence
    from typing import Any
    from typing import TypeVar

    from glotaran.io.interface import DataIoInterface
    from glotaran.io.interface import ProjectIoInterface
    from glotaran.model.megacomplex import Megacomplex

    _PluginType = TypeVar("_PluginType", type[Megacomplex], DataIoInterface, ProjectIoInterface)
    _PluginInstantiableType = TypeVar(
        "_PluginInstantiableType", DataIoInterface, ProjectIoInterface
    )

    GenericPluginInstance = TypeVar("GenericPluginInstance", bound=object)


class __PluginRegistry:
    """Central Plugin Registry.

    This is super private since if anyone messes with it, the pluginsystem could break.
    """

    megacomplex: MutableMapping[str, type[Megacomplex]] = {}
    data_io: MutableMapping[str, DataIoInterface] = {}
    project_io: MutableMapping[str, ProjectIoInterface] = {}


def full_plugin_name(plugin: object | type[object]) -> str:
    """Full name of a plugin instance/class similar to the ``repr``.

    Parameters
    ----------
    plugin : object | type[object]
        plugin instance/class

    Examples
    --------
    >>> from glotaran.builtin.io.sdt.sdt_file_reader import SdtDataIo
    >>> full_plugin_name(SdtDataIo)
    "glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo"
    >>> full_plugin_name(SdtDataIo("sdt"))
    "glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo"

    Returns
    -------
    str
        Full name of the plugin.
    """
    if isinstance(plugin, type):
        return f"{plugin.__module__}.{plugin.__name__}"
    else:
        return f"{plugin.__module__}.{type(plugin).__name__}"


class PluginOverwriteWarning(UserWarning):
    """Warning used if a plugin tries to overwrite and existing plugin."""

    def __init__(
        self,
        *args: Any,
        old_key: str,
        old_plugin: object | type[object],
        new_plugin: object | type[object],
        plugin_set_func_name: str,
    ):
        """Use old and new plugin and keys to give verbose warning message.

        Parameters
        ----------
        old_key : str
            Old registry key.
        old_plugin :  object | type[object]
            Old plugin ('registry[old_key]').
        new_plugin :  object | type[object]
            New Plugin ('registry[new_key]').
        plugin_set_func_name: str
            Name of the function used to pin a plugin.
        *args : Any
            Additional args passed to the super constructor.
        """
        old_plugin_name = full_plugin_name(old_plugin)
        new_plugin_name = full_plugin_name(new_plugin)
        message = (
            f"The plugin '{new_plugin_name}' tried to overwrite the plugin '{old_plugin_name}', "
            f"with the access_name {old_key!r}. "
            f"Use {plugin_set_func_name}({old_key!r}, {new_plugin_name!r}) "
            f"to use {new_plugin_name!r} instead."
        )
        super().__init__(message, *args)


def load_plugins():
    """Initialize plugins registered under the entrypoint 'glotaran.plugins'.

    For an entry_point to be considered a glotaran plugin it just needs to start with
    'glotaran.plugins', which allows for an easy extendability.

    Currently used builtin entrypoints are:

    - ``glotaran.plugins.data_io``
    - ``glotaran.plugins.megacomplex``
    - ``glotaran.plugins.project_io``
    """
    if "DEACTIVATE_GTA_PLUGINS" not in os.environ:  # pragma: no branch
        for entry_point_name, entry_points in metadata.entry_points().items():
            if entry_point_name.startswith("glotaran.plugins"):
                for entry_point in entry_points:
                    entry_point.load()


def set_plugin(
    plugin_register_key: str,
    full_plugin_name: str,
    plugin_registry: MutableMapping[str, _PluginType],
    plugin_register_key_name: str = "format_name",
) -> None:
    """Set a plugins short name to a specific plugin referred by its full name.

    This can be used to ensure that a specific plugin is used in case there
    are conflicting plugins installed.

    Parameters
    ----------
    plugin_register_key : str
        Name of the plugin under which it is registered.
    full_plugin_name : str
        Full name (import path) of the registered plugin.
    plugin_registry : MutableMapping[str, _PluginType]
        Registry the plugin should be set in to.
    plugin_register_key_name: str
        Name of the arg passed ``plugin_register_key`` in the function that implements
        ``set_plugin``.

    Raises
    ------
    ValueError
        If ``plugin_register_key`` has the character '.' in it.
    ValueError
        If there isn't a registered plugin with the key ``full_plugin_name``.

    See Also
    --------
    add_plugin_to_registry
    full_plugin_name
    """
    if "." in plugin_register_key:
        raise ValueError(
            f"The value of {plugin_register_key_name!r} isn't "
            "allowed to contain the character '.' ."
        )
    if "." not in full_plugin_name or not is_registered_plugin(
        plugin_register_key=full_plugin_name, plugin_registry=plugin_registry
    ):
        known_plugins = list(
            filter(lambda plugin_name: "." in plugin_name, plugin_registry.keys())
        )
        raise ValueError(
            f"There isn't a plugin registered under the full name {full_plugin_name!r}.\n"
            f"Maybe you need to install a plugin? Known plugins are:\n {known_plugins}"
        )
    plugin_registry[plugin_register_key] = plugin_registry[full_plugin_name]


def add_plugin_to_registry(
    plugin_register_key: str,
    plugin: _PluginType,
    plugin_registry: MutableMapping[str, _PluginType],
    plugin_set_func_name: str,
    instance_identifier: str = "",
) -> None:
    """Add a plugin with name ``plugin_register_key`` to the given registry.

    In addition it also adds the plugin with it full import path name as key,
    which allows for a better reproducibility in case there are conflicting plugins.

    Parameters
    ----------
    plugin_register_key : str
        Name of the plugin under which it is registered.
    plugin: _PluginType
        Plugin to be added to the registry.
    plugin_registry: MutableMapping[str, _PluginType]
        Registry the plugin should be added to.
    plugin_set_func_name: str
        Name of the function used to pin a plugin.
    instance_identifier: str
        Used to differentiate between plugin instances
        (e.g. different format for IO plugins)

    Raises
    ------
    ValueError
        If ``plugin_register_key`` has the character '.' in it.

    See Also
    --------
    add_instantiated_plugin_to_register
    full_plugin_name
    """
    if "." in plugin_register_key:
        raise ValueError(
            "The character '.' isn't allowed in the name of a plugin, "
            f"you provided the name {plugin_register_key!r}."
        )
    if plugin_register_key in plugin_registry:
        old_key = plugin_register_key
        plugin_register_key = full_plugin_name(plugin)
        if full_plugin_name(plugin_registry[old_key]) != full_plugin_name(plugin):
            warn(
                PluginOverwriteWarning(
                    old_key=old_key,
                    old_plugin=plugin_registry[old_key],
                    new_plugin=plugin,
                    plugin_set_func_name=plugin_set_func_name,
                ),
                stacklevel=4,
            )
    if instance_identifier:
        instance_identifier = f"_{instance_identifier}"
    plugin_registry[f"{full_plugin_name(plugin)}{instance_identifier}"] = plugin
    plugin_registry[plugin_register_key] = plugin


def add_instantiated_plugin_to_registry(
    plugin_register_keys: str | list[str],
    plugin_class: type[_PluginInstantiableType],
    plugin_registry: MutableMapping[str, _PluginInstantiableType],
    plugin_set_func_name: str,
) -> None:
    """Add instances of plugin_class to the given registry.

    Parameters
    ----------
    plugin_register_keys : str | list[str]
        Name/-s of the plugin under which it is registered.
    plugin_class : type[_PluginInstantiableType]
        Pluginclass which should be instantiated with ``plugin_register_keys``
        and added to the registry.
    plugin_registry : MutableMapping[str, _PluginInstantiableType]
        Registry the plugin should be added to.
    plugin_set_func_name: str
        Name of the function used to pin a plugin.

    See Also
    --------
    add_plugin_to_register
    """
    if isinstance(plugin_register_keys, str):
        plugin_register_keys = [plugin_register_keys]
    for plugin_register_key in plugin_register_keys:
        add_plugin_to_registry(
            plugin_register_key=plugin_register_key,
            plugin=plugin_class(plugin_register_key),
            plugin_registry=plugin_registry,
            plugin_set_func_name=plugin_set_func_name,
            instance_identifier=plugin_register_key,
        )


def registered_plugins(
    plugin_registry: MutableMapping[str, _PluginType], full_names: bool = False
) -> list[str]:
    """Names of the plugins in the given registry.

    Parameters
    ----------
    plugin_registry : MutableMapping[str, _PluginType]
        Registry to search in.
    full_names: bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    list[str]
        List of plugin names in plugin_registry.
    """
    if full_names:
        return sorted(plugin_registry.keys())

    else:
        return sorted(filter(lambda key: "." not in key, plugin_registry.keys()))


def is_registered_plugin(
    plugin_register_key: str, plugin_registry: MutableMapping[str, _PluginType]
) -> bool:
    """Check if a plugin with name ``plugin_register_key`` is registered in the given registry.

    Parameters
    ----------
    plugin_register_key : str
        Name of the plugin under which it is registered.
    plugin_registry : MutableMapping[str, _PluginType]
        Registry to search in.

    Returns
    -------
    bool
        Whether or not a plugin is in the registry.
    """
    return plugin_register_key in plugin_registry


def get_plugin_from_registry(
    plugin_register_key: str,
    plugin_registry: MutableMapping[str, _PluginType],
    not_found_error_message: str,
) -> _PluginType:
    """Retrieve a plugin with name ``plugin_register_key`` is registered in a given registry.

    Parameters
    ----------
    plugin_register_key : str
        Name of the plugin under which it is registered.
    plugin_registry : MutableMapping[str, _PluginType]
        Registry to search in.
    not_found_error_message : str
        Error message to be shown if the plugin wasn't found.

    Returns
    -------
    _PluginType
        Plugin from the plugin Registry.

    Raises
    ------
    ValueError
        If there was no plugin registered under the name ``plugin_register_key``.
    """
    if not is_registered_plugin(plugin_register_key, plugin_registry):
        raise ValueError(not_found_error_message)
    else:
        return plugin_registry[plugin_register_key]


def get_method_from_plugin(
    plugin: object | type[object],
    method_name: str,
) -> Callable[..., Any]:
    """Retrieve a method callabe from an class or instance plugin.

    Parameters
    ----------
    plugin : object | type[object],
        Plugin instance or class.
    method_name : str
        Method name, e.g. load_megacomplex.

    Returns
    -------
    Callable[..., Any]
        Method callable.

    Raises
    ------
    ValueError
        If plugin has an attribute with that name but it isn't callable.
    ValueError
        If plugin misses the attribute.
    """
    not_a_method_error_message = (
        f"The plugin {full_plugin_name(plugin)!r} has no method {method_name!r}"
    )
    try:
        possible_method = getattr(plugin, method_name)
        if callable(possible_method):
            return possible_method
        else:
            raise ValueError(not_a_method_error_message)
    except AttributeError as err:
        raise ValueError(not_a_method_error_message) from err


def show_method_help(
    plugin: object | type[object],
    method_name: str,
) -> None:
    """Show help on a method as if it was called directly on it.

    Parameters
    ----------
    plugin : object | type[object],
        Plugin instance or class.
    method_name : str
        Method name, e.g. load_megacomplex.
    """
    method = get_method_from_plugin(plugin, method_name)
    help(method)


def methods_differ_from_baseclass(
    method_names: str | Sequence[str],
    plugin: GenericPluginInstance | type[GenericPluginInstance],
    base_class: type[GenericPluginInstance],
) -> Generator[bool, None, None]:
    """Check if a plugins methods implementation differ from its baseclass.

    Based on the assumption that ``base_class`` didn't implement the methods
    (e.g. :class:`DataIoInterface` or :class:`ProjectIoInterface`), this can be
    used to to create a 'supported methods' list.

    Parameters
    ----------
    method_names : str | list[str]
        Name|s of the method|s
    plugin : GenericPluginInstance | type[GenericPluginInstance]
        Plugin class or instance.
    base_class : type[GenericPluginInstance]
        Base class the plugin inherited from.

    Yields
    ------
    bool
        Whether or not a plugins method differs from the implementation in ``base_class``.
    """
    if isinstance(method_names, str):
        method_names = [method_names]
    for method_name in method_names:
        plugin_method = get_method_from_plugin(plugin, method_name)
        base_class_method = get_method_from_plugin(base_class, method_name)
        yield plugin_method.__code__ != base_class_method.__code__


def methods_differ_from_baseclass_table(
    method_names: str | Sequence[str],
    plugin_registry_keys: str | Sequence[str],
    get_plugin_function: Callable[[str], GenericPluginInstance | type[GenericPluginInstance]],
    base_class: type[GenericPluginInstance],
    plugin_names: bool = False,
) -> Generator[list[str | bool], None, None]:
    """Create table of which plugins methods differ from their baseclass.

    This uses the assumption that all plugins have the same ``base_class``.

    The main purpose of this function is to show the user which plugin implements
    which methods differently than its baseclass.

    Based on the assumption that ``base_class`` didn't implement the methods
    (e.g. :class:`DataIoInterface` or :class:`ProjectIoInterface`), this can be
    used to to create a 'supported methods' table.

    Parameters
    ----------
    method_names : str | list[str]
        Name|s of the method|s.
    plugin_registry_keys : str | list[str]
        Keys the plugins are registered under
        (e.g. return value of the implementation of func:`registered_plugins`)
    get_plugin_function: Callable[[str], GenericPluginInstance | type[GenericPluginInstance]]
        Function to get plugin from plugin registry.
    base_class : type[GenericPluginInstance]
        Base class the plugin inherited from.
    plugin_names : bool
        Whether or not to add the names of the plugins to the lists.

    Yields
    ------
    list[str | bool]
        Row with the first value being the ``plugin_registry_key`` and the others whether or not
        a plugins method differs from ``base_class``.

    See Also
    --------
    methods_differ_from_baseclass
    """
    if isinstance(plugin_registry_keys, str):
        plugin_registry_keys = [plugin_registry_keys]
    for plugin_registry_key in plugin_registry_keys:
        plugin = get_plugin_function(plugin_registry_key)
        differs_list = methods_differ_from_baseclass(method_names, plugin, base_class)
        row: list[str | bool] = [f"`{plugin_registry_key}`", *differs_list]
        if plugin_names:
            if isinstance(plugin, type):
                row.append(f"`{full_plugin_name(plugin)}`")
            elif "." in plugin_registry_key:
                row.append(f"`{plugin_registry_key}`")
            else:
                row.append(f"`{full_plugin_name(plugin)}_{plugin_registry_key}`")
        yield row


def supported_file_extensions(
    method_names: str | Sequence[str],
    plugin_registry_keys: str | Sequence[str],
    get_plugin_function: Callable[[str], GenericPluginInstance | type[GenericPluginInstance]],
    base_class: type[GenericPluginInstance],
) -> Generator[str, None, None]:
    """Get file extensions for plugins that support all methods in ``method_names``.

    Parameters
    ----------
    method_names : str | list[str]
        Name|s of the method|s.
    plugin_registry_keys : str | list[str]
        Keys the plugins are registered under
        (e.g. return value of the implementation of func:`registered_plugins`)
    get_plugin_function: Callable[[str], GenericPluginInstance | type[GenericPluginInstance]]
        Function to get plugin from plugin registry.
    base_class : type[GenericPluginInstance]
        Base class the plugin inherited from.

    Yields
    ------
    Generator[str, None, None]
        File extension supported by all methods in ``method_names``.

    See Also
    --------
    methods_differ_from_baseclass
    methods_differ_from_baseclass_table
    """
    for plugin_registry_key, *differs_list in methods_differ_from_baseclass_table(
        method_names, plugin_registry_keys, get_plugin_function, base_class
    ):
        format_name_str: str = cast(str, plugin_registry_key).replace("`", "")
        if format_name_str.endswith("_str"):
            continue
        if all(cast(Iterable[bool], differs_list)) is True:
            yield f".{format_name_str}"



"""Glotaran types module containing commonly used types."""
from collections.abc import Mapping
from collections.abc import Sequence
from pathlib import Path
from typing import TypeVar
from typing import Union

import numpy as np

try:
    from numpy._typing._array_like import _SupportsArray
except ImportError:
    # numpy < 1.23
    from numpy.typing._array_like import _SupportsArray  # type:ignore[no-redef]

import xarray as xr

T = TypeVar("T")
StrOrPath = Union[str, Path]
LoadableDataset = Union[StrOrPath, xr.Dataset, xr.DataArray]
DatasetMappable = Union[LoadableDataset, Sequence[LoadableDataset], Mapping[str, LoadableDataset]]


ArrayLike = np.ndarray



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import Model
from glotaran.model import ParameterType
from glotaran.model import item
from glotaran.model import megacomplex

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@megacomplex()
class SimpleTestMegacomplex(Megacomplex):
    type: str = "simple-test-mc"
    dimension: str = "model"
    is_index_dependent: bool

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        compartments = ["s1", "s2"]
        array = np.zeros((model_axis.size, len(compartments)))

        for i in range(len(compartments)):
            for j in range(model_axis.size):
                array[j, i] = (i + j) * model_axis[j]
        if self.is_index_dependent:
            array = np.array([array] * global_axis.size)
        return compartments, array

    def finalize_data(
        self,
        dataset_model,
        dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        # not needed in the tests
        pass


SimpleTestModel = Model.create_class_from_megacomplexes([SimpleTestMegacomplex])


@item
class SimpleDatasetModel(DatasetModel):
    kinetic: list[ParameterType]


@megacomplex(dataset_model_type=SimpleDatasetModel)
class SimpleKineticMegacomplex(Megacomplex):
    type: str = "simple-kinetic-test-mc"
    dimension: str = "model"
    is_index_dependent: bool

    def calculate_matrix(
        self,
        dataset_model,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        kinpar = -1 * np.asarray(dataset_model.kinetic)
        if dataset_model.label == "dataset3":
            # this case is for the ThreeDatasetDecay test
            compartments = [f"s{i+2}" for i in range(len(kinpar))]
        else:
            compartments = [f"s{i+1}" for i in range(len(kinpar))]
        array = np.exp(np.outer(model_axis, kinpar))
        if self.is_index_dependent:
            array = np.array([array] * global_axis.size)
        return compartments, array

    def finalize_data(
        self,
        dataset_model,
        dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        pass


@megacomplex()
class SimpleSpectralMegacomplex(Megacomplex):
    type: str = "simple-spectral-test-mc"
    dimension: str = "global"

    def calculate_matrix(
        self,
        dataset_model,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        kinpar = dataset_model.kinetic
        if dataset_model.label == "dataset3":
            # this case is for the ThreeDatasetDecay test
            compartments = [f"s{i+2}" for i in range(len(kinpar))]
        else:
            compartments = [f"s{i+1}" for i in range(len(kinpar))]
        array = np.asarray([[1 for _ in range(model_axis.size)] for _ in compartments]).T
        return compartments, array


@megacomplex()
class ShapedSpectralMegacomplex(Megacomplex):
    type: str = "shaped-spectral-test-mc"
    dimension: str = "global"
    location: list[ParameterType]
    amplitude: list[ParameterType]
    delta: list[ParameterType]

    def calculate_matrix(
        self,
        dataset_model,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        location = np.asarray(self.location)
        amp = np.asarray(self.amplitude)
        delta = np.asarray(self.delta)

        array = np.empty((location.size, model_axis.size), dtype=np.float64)

        for i in range(location.size):
            array[i, :] = amp[i] * np.exp(
                -np.log(2) * np.square(2 * (model_axis - location[i]) / delta[i])
            )
        compartments = [f"s{i+1}" for i in range(location.size)]
        return compartments, array.T

    def finalize_data(
        self,
        dataset_model,
        dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        pass


DecayModel = Model.create_class_from_megacomplexes(
    [SimpleKineticMegacomplex, SimpleSpectralMegacomplex, ShapedSpectralMegacomplex]
)



import numpy as np
import pytest
import xarray as xr

from glotaran.optimization.data_provider import DataProvider
from glotaran.optimization.data_provider import DataProviderLinked
from glotaran.optimization.matrix_provider import MatrixProviderLinked
from glotaran.optimization.matrix_provider import MatrixProviderUnlinked
from glotaran.optimization.test.models import SimpleTestModel
from glotaran.parameter import Parameters
from glotaran.project import Scheme


@pytest.fixture()
def dataset_one() -> xr.Dataset:
    global_axis = [1, 5, 6]
    model_axis = [5, 7, 9, 12]

    data = xr.DataArray(
        np.ones((4, 3)), coords=[("model", model_axis), ("global", global_axis)]
    ).to_dataset(name="data")

    data["weight"] = xr.ones_like(data.data) * 0.5
    return data


@pytest.fixture()
def dataset_two() -> xr.Dataset:
    global_axis = [0, 3, 7, 10]
    model_axis = [4, 11, 15]

    return xr.DataArray(
        np.ones((4, 3)) * 2, coords=[("global", global_axis), ("model", model_axis)]
    ).to_dataset(name="data")


@pytest.fixture()
def scheme(dataset_one: xr.Dataset, dataset_two: xr.Dataset) -> Scheme:
    model = SimpleTestModel(
        **{
            "megacomplex": {"m1": {"type": "simple-test-mc", "is_index_dependent": False}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                },
                "dataset2": {
                    "megacomplex": ["m1"],
                },
            },
        }
    )
    parameters = Parameters.from_list([])

    data = {"dataset1": dataset_one, "dataset2": dataset_two}
    return Scheme(model, parameters, data, clp_link_tolerance=1)


def test_matrix_provider_unlinked_index_independent(scheme: Scheme):
    dataset_group = scheme.model.get_dataset_groups()["default"]
    dataset_group.set_parameters(scheme.parameters)
    data_provider = DataProvider(scheme, dataset_group)
    matrix_provider = MatrixProviderUnlinked(dataset_group, data_provider)
    matrix_provider.calculate()
    _, matrices = matrix_provider.get_result()

    assert "dataset1" in matrices
    assert matrices["dataset1"].shape == (scheme.data["dataset1"].model.size, 2)
    assert all(matrices["dataset1"].clp_label == ["s1", "s2"])

    assert "dataset2" in matrices
    assert matrices["dataset2"].shape == (scheme.data["dataset2"].model.size, 2)
    assert all(matrices["dataset2"].clp_label == ["s1", "s2"])

    # 2 compartments * 3 items in global axis of dataset1
    # + 2 compartments * 4 items in global axis of dataset2
    assert matrix_provider.number_of_clps == (2 * 3) + (2 * 4)


def test_matrix_provider_linked_index_independent(scheme: Scheme):
    dataset_group = scheme.model.get_dataset_groups()["default"]
    dataset_group.set_parameters(scheme.parameters)
    data_provider = DataProviderLinked(scheme, dataset_group)
    matrix_provider = MatrixProviderLinked(dataset_group, data_provider)
    matrix_provider.calculate()
    _, matrices = matrix_provider.get_result()

    dataset1_size = scheme.data["dataset1"].coords["model"].size
    dataset2_size = scheme.data["dataset2"].coords["model"].size

    assert "dataset1" in matrices
    assert matrices["dataset1"].shape == (dataset1_size, 2)
    assert all(matrices["dataset1"].clp_label == ["s1", "s2"])

    assert "dataset2" in matrices
    assert matrices["dataset2"].shape == (dataset2_size, 2)
    assert all(matrices["dataset2"].clp_label == ["s1", "s2"])

    assert matrix_provider.get_aligned_matrix_container(0).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(1).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(2).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(3).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(4).clp_labels == ["s1", "s2"]

    assert matrix_provider.get_aligned_matrix_container(0).matrix.shape == (
        dataset1_size + dataset2_size,
        2,
    )
    assert matrix_provider.get_aligned_matrix_container(1).matrix.shape == (dataset2_size, 2)
    assert matrix_provider.get_aligned_matrix_container(2).matrix.shape == (dataset1_size, 2)
    assert matrix_provider.get_aligned_matrix_container(3).matrix.shape == (
        dataset1_size + dataset2_size,
        2,
    )
    assert matrix_provider.get_aligned_matrix_container(4).matrix.shape == (dataset2_size, 2)

    # 2 compartments * 5 items in aligned global axis
    # See also: test_data_provider_linking_methods
    assert matrix_provider.number_of_clps == 2 * 5


def test_matrix_provider_unlinked_index_dependent(scheme: Scheme):
    scheme.model.megacomplex["m1"].is_index_dependent = True  # type:ignore[attr-defined]
    dataset_group = scheme.model.get_dataset_groups()["default"]
    dataset_group.set_parameters(scheme.parameters)
    data_provider = DataProvider(scheme, dataset_group)
    matrix_provider = MatrixProviderUnlinked(dataset_group, data_provider)
    matrix_provider.calculate()
    _, matrices = matrix_provider.get_result()

    dataset1_model_size = scheme.data["dataset1"].coords["model"].size
    dataset1_global_size = scheme.data["dataset1"].coords["global"].size
    dataset2_model_size = scheme.data["dataset2"].coords["model"].size
    dataset2_global_size = scheme.data["dataset2"].coords["global"].size

    assert "dataset1" in matrices
    assert matrices["dataset1"].shape == (dataset1_global_size, dataset1_model_size, 2)
    assert all(matrices["dataset1"].clp_label == ["s1", "s2"])

    assert "dataset2" in matrices
    assert matrices["dataset2"].shape == (dataset2_global_size, dataset2_model_size, 2)
    assert all(matrices["dataset2"].clp_label == ["s1", "s2"])

    # 2 compartments * 3 items in global axis of dataset1
    # + 2 compartments * 4 items in global axis of dataset2
    assert matrix_provider.number_of_clps == (2 * 3) + (2 * 4)


def test_matrix_provider_linked_index_dependent(scheme: Scheme):
    scheme.model.megacomplex["m1"].is_index_dependent = True  # type:ignore[attr-defined]
    dataset_group = scheme.model.get_dataset_groups()["default"]
    dataset_group.set_parameters(scheme.parameters)
    data_provider = DataProviderLinked(scheme, dataset_group)
    matrix_provider = MatrixProviderLinked(dataset_group, data_provider)
    matrix_provider.calculate()
    _, matrices = matrix_provider.get_result()

    dataset1_model_size = scheme.data["dataset1"].coords["model"].size
    dataset1_global_size = scheme.data["dataset1"].coords["global"].size
    dataset2_model_size = scheme.data["dataset2"].coords["model"].size
    dataset2_global_size = scheme.data["dataset2"].coords["global"].size

    assert "dataset1" in matrices
    assert matrices["dataset1"].shape == (dataset1_global_size, dataset1_model_size, 2)
    assert all(matrices["dataset1"].clp_label == ["s1", "s2"])

    assert "dataset2" in matrices
    assert matrices["dataset2"].shape == (dataset2_global_size, dataset2_model_size, 2)
    assert all(matrices["dataset2"].clp_label == ["s1", "s2"])

    assert matrix_provider.get_aligned_matrix_container(0).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(1).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(2).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(3).clp_labels == ["s1", "s2"]
    assert matrix_provider.get_aligned_matrix_container(4).clp_labels == ["s1", "s2"]

    assert matrix_provider.get_aligned_matrix_container(0).matrix.shape == (
        dataset1_model_size + dataset2_model_size,
        2,
    )
    assert matrix_provider.get_aligned_matrix_container(1).matrix.shape == (dataset2_model_size, 2)
    assert matrix_provider.get_aligned_matrix_container(2).matrix.shape == (dataset1_model_size, 2)
    assert matrix_provider.get_aligned_matrix_container(3).matrix.shape == (
        dataset1_model_size + dataset2_model_size,
        2,
    )
    assert matrix_provider.get_aligned_matrix_container(4).matrix.shape == (dataset2_model_size, 2)

    # 2 compartments * 5 items in aligned global axis
    # See also: test_data_provider_linking_methods
    assert matrix_provider.number_of_clps == 2 * 5



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from glotaran.builtin.megacomplexes.decay.test.test_decay_megacomplex import DecayModel
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.model import Model
from glotaran.model import fill_item
from glotaran.optimization.matrix_provider import MatrixProvider
from glotaran.optimization.optimize import optimize
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate

SpectralModel = Model.create_class_from_megacomplexes([SpectralMegacomplex])


class OneCompartmentModelInvertedAxis:
    decay_model = DecayModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"type": "decay", "k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    decay_parameters = Parameters.from_list([101e-4, [1, {"vary": False, "non-negative": False}]])

    spectral_model = SpectralModel(
        **{
            "megacomplex": {
                "mc1": {"type": "spectral", "shape": {"s1": "sh1"}},
            },
            "shape": {
                "sh1": {
                    "type": "gaussian",
                    "amplitude": "1",
                    "location": "2",
                    "width": "3",
                }
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["mc1"],
                    "spectral_axis_scale": 1e7,
                    "spectral_axis_inverted": True,
                },
            },
        }
    )

    spectral_parameters = Parameters.from_list([7, 1e7 / 10000, 800, -1])

    time = np.arange(-10, 50, 1.5)
    spectral = np.arange(5000, 15000, 20)
    axis = {"time": time, "spectral": spectral}

    decay_dataset_model = fill_item(decay_model.dataset["dataset1"], decay_model, decay_parameters)
    matrix = MatrixProvider.calculate_dataset_matrix(decay_dataset_model, spectral, time)
    decay_compartments = matrix.clp_labels
    clp = xr.DataArray(matrix.matrix, coords=[("time", time), ("clp_label", decay_compartments)])


class OneCompartmentModelNegativeSkew:
    decay_model = DecayModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"type": "decay", "k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    decay_parameters = Parameters.from_list([101e-4, [1, {"vary": False, "non-negative": False}]])

    spectral_model = SpectralModel(
        **{
            "megacomplex": {
                "mc1": {"type": "spectral", "shape": {"s1": "sh1"}},
            },
            "shape": {
                "sh1": {
                    "type": "skewed-gaussian",
                    "location": "1",
                    "width": "2",
                    "skewness": "3",
                }
            },
            "dataset": {
                "dataset1": {"megacomplex": ["mc1"], "spectral_axis_scale": 1},
            },
        }
    )

    spectral_parameters = Parameters.from_list([500, 80, -1])

    time = np.arange(-10, 50, 1.5)
    spectral = np.arange(400, 600, 5)
    axis = {"time": time, "spectral": spectral}

    decay_dataset_model = fill_item(decay_model.dataset["dataset1"], decay_model, decay_parameters)
    matrix = MatrixProvider.calculate_dataset_matrix(decay_dataset_model, spectral, time)
    decay_compartments = matrix.clp_labels
    clp = xr.DataArray(matrix.matrix, coords=[("time", time), ("clp_label", decay_compartments)])


class OneCompartmentModelPositivSkew(OneCompartmentModelNegativeSkew):
    spectral_parameters = Parameters.from_list([500, 80, 1])


class OneCompartmentModelZeroSkew(OneCompartmentModelNegativeSkew):
    spectral_parameters = Parameters.from_list([500, 80, 0])


class ThreeCompartmentModel:
    decay_model = DecayModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1", "s2", "s3"], "parameters": ["4", "4", "4"]},
            },
            "megacomplex": {
                "mc1": {"type": "decay", "k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                        ("s2", "s2"): "2",
                        ("s3", "s3"): "3",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    decay_parameters = Parameters.from_list(
        [101e-4, 101e-5, 101e-6, [1, {"vary": False, "non-negative": False}]]
    )

    spectral_model = SpectralModel(
        **{
            "megacomplex": {
                "mc1": {
                    "type": "spectral",
                    "shape": {
                        "s1": "sh1",
                        "s2": "sh2",
                        "s3": "sh3",
                    },
                },
            },
            "shape": {
                "sh1": {
                    "type": "gaussian",
                    "amplitude": "1",
                    "location": "2",
                    "width": "3",
                },
                "sh2": {
                    "type": "gaussian",
                    "amplitude": "4",
                    "location": "5",
                    "width": "6",
                },
                "sh3": {
                    "type": "gaussian",
                    "amplitude": "7",
                    "location": "8",
                    "width": "9",
                },
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    spectral_parameters = Parameters.from_list(
        [
            7,
            450,
            80,
            20,
            550,
            50,
            10,
            580,
            10,
        ]
    )

    time = np.arange(-10, 50, 1.5)
    spectral = np.arange(400, 600, 5)
    axis = {"time": time, "spectral": spectral}

    decay_dataset_model = fill_item(decay_model.dataset["dataset1"], decay_model, decay_parameters)
    matrix = MatrixProvider.calculate_dataset_matrix(decay_dataset_model, spectral, time)
    decay_compartments = matrix.clp_labels
    clp = xr.DataArray(matrix.matrix, coords=[("time", time), ("clp_label", decay_compartments)])


@pytest.mark.parametrize(
    "suite",
    [
        OneCompartmentModelNegativeSkew,
        OneCompartmentModelPositivSkew,
        OneCompartmentModelZeroSkew,
        ThreeCompartmentModel,
    ],
)
def test_spectral_model(suite):
    model = suite.spectral_model
    print(model.validate())
    assert model.valid()

    wanted_parameters = suite.spectral_parameters
    print(model.validate(wanted_parameters))
    print(wanted_parameters)
    assert model.valid(wanted_parameters)

    initial_parameters = suite.spectral_parameters
    print(model.validate(initial_parameters))
    assert model.valid(initial_parameters)

    print(model.markdown(initial_parameters))

    dataset = simulate(model, "dataset1", wanted_parameters, suite.axis, suite.clp)
    # possible test improvement: noise=True, noise_std_dev=1e-8, noise_seed=123

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["spectral"].size)

    data = {"dataset1": dataset}

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for param in result.optimized_parameters.all():
        assert np.allclose(param.value, wanted_parameters.get(param.label).value, rtol=1e-1)
        # should probably change rtol -> atol

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["spectral"], resultdata["spectral"])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, rtol=1e-2)
    # should probably change rtol -> atol
    assert "species_associated_concentrations" in resultdata
    assert resultdata.species_associated_concentrations.shape == (
        suite.axis["time"].size,
        len(suite.decay_compartments),
    )
    assert "species_spectra" in resultdata
    assert resultdata.species_spectra.shape == (
        suite.axis["spectral"].size,
        len(suite.decay_compartments),
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.decay.initial_concentration import InitialConcentration
from glotaran.builtin.megacomplexes.decay.irf import Irf
from glotaran.builtin.megacomplexes.decay.k_matrix import KMatrix
from glotaran.builtin.megacomplexes.decay.util import calculate_matrix
from glotaran.builtin.megacomplexes.decay.util import finalize_data
from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import ModelItemType
from glotaran.model import item
from glotaran.model import megacomplex

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@item
class DecayDatasetModel(DatasetModel):
    initial_concentration: ModelItemType[InitialConcentration] | None = None
    irf: ModelItemType[Irf] | None = None


@megacomplex(dataset_model_type=DecayDatasetModel)
class DecayMegacomplex(Megacomplex):
    dimension: str = "time"
    type: str = "decay"
    k_matrix: list[ModelItemType[KMatrix]]

    def get_compartments(self, dataset_model: DatasetModel) -> list[str]:
        if dataset_model.initial_concentration is None:
            raise ModelError(
                f'No initial concentration specified in dataset "{dataset_model.label}"'
            )
        return [
            compartment
            for compartment in dataset_model.initial_concentration.compartments
            if compartment in self.get_k_matrix().involved_compartments()
        ]

    def get_initial_concentration(
        self, dataset_model: DatasetModel, normalized: bool = True
    ) -> np.ndarray:
        compartments = self.get_compartments(dataset_model)
        idx = [
            compartment in compartments
            for compartment in dataset_model.initial_concentration.compartments
        ]
        initial_concentration = (
            dataset_model.initial_concentration.normalized()
            if normalized
            else np.asarray(dataset_model.initial_concentration.parameters)
        )
        return initial_concentration[idx]

    def get_k_matrix(self) -> KMatrix:
        full_k_matrix = None
        for k_matrix in self.k_matrix:
            if full_k_matrix is None:
                full_k_matrix = k_matrix
            # If multiple k matrices are present, we combine them
            else:
                full_k_matrix = full_k_matrix.combine(k_matrix)
        return full_k_matrix

    def get_a_matrix(self, dataset_model: DatasetModel) -> np.ndarray:
        return self.get_k_matrix().a_matrix(
            self.get_compartments(dataset_model), self.get_initial_concentration(dataset_model)
        )

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        return calculate_matrix(self, dataset_model, global_axis, model_axis, **kwargs)

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        finalize_data(dataset_model, dataset, is_full_model, as_global)



"""The glotaran model registry module."""
from __future__ import annotations

from pathlib import Path

from glotaran.io import load_model
from glotaran.plugin_system.project_io_registration import supported_file_extensions_project_io
from glotaran.project.generators.generator import GeneratorArguments
from glotaran.project.generators.generator import generate_model_yml
from glotaran.project.project_registry import ProjectRegistry


class ProjectModelRegistry(ProjectRegistry):
    """A registry for models."""

    def __init__(self, directory: Path):
        """Initialize a model registry.

        Parameters
        ----------
        directory : Path
            The registry directory.
        """
        super().__init__(
            directory / "models",
            supported_file_extensions_project_io("load_model"),
            load_model,
            item_name="Model",
        )

    def generate_model(
        self,
        name: str,
        generator_name: str,
        generator_arguments: GeneratorArguments,
        *,
        allow_overwrite: bool = False,
        ignore_existing: bool = False,
    ):
        """Generate a model.

        Parameters
        ----------
        name : str
            The name of the model.
        generator_name : str
            The generator for the model.
        generator_arguments : GeneratorArguments
            Arguments for the generator.
        allow_overwrite: bool
            Whether to overwrite an existing model.
        ignore_existing: bool
            Whether to ignore generation of a model file if it already exists.

        Raises
        ------
        FileExistsError
            Raised if model is already existing and `allow_overwrite=False`.
        """
        model_yml = generate_model_yml(
            generator_name=generator_name, generator_arguments=generator_arguments
        )
        model_file = self._directory / f"{name}.yml"
        if model_file.exists() and ignore_existing:
            return

        if model_file.exists() and not allow_overwrite:
            raise FileExistsError(f"Model {name!r} already exists and `allow_overwrite=False`.")

        model_file.write_text(model_yml)



"""Module containing context manager to capture stdout output and still forward it to stdout."""
from __future__ import annotations

import sys
from io import StringIO
from types import TracebackType


class TeeContext:
    """Context manager that allows to work with string written to stdout.

    This context manager behaves similar to the ``tee`` shell command.
    https://linuxize.com/post/linux-tee-command
    """

    def __init__(self) -> None:
        """Create new ``StringIO`` buffer and save reference to original ``sys.stdout``."""
        self.buffer = StringIO()
        self.stdout = sys.stdout

    def __enter__(self) -> TeeContext:
        """Replace ``sys.stdout`` on entering the context.

        Returns
        -------
        TeeContext
            Instance that can be read from.
        """
        sys.stdout = self  # type:ignore[assignment]
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        """Restore ``sys.stdout`` on exiting the context."""
        sys.stdout = self.stdout
        return None

    def write(self, data: str) -> None:
        """Write to buffer and original ``sys.stdout``.

        Parameters
        ----------
        data: str
            String to write to stdout and buffer.
        """
        self.buffer.write(data)
        self.stdout.write(data)

    def read(self) -> str:
        """Return values written to buffer.

        Returns
        -------
        str
            Text written to buffer.
        """
        return self.buffer.getvalue()

    def flush(self) -> None:
        """Flush values in the buffer."""
        self.buffer.flush()



"""The parameter class."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING
from typing import Any

import asteval
import numpy as np
from attr import ib
from attrs import Attribute
from attrs import asdict
from attrs import define
from attrs import evolve
from attrs import fields
from attrs import filters
from attrs import validators

from glotaran.typing.types import _SupportsArray
from glotaran.utils.attrs_helper import no_default_vals_in_repr
from glotaran.utils.helpers import nan_or_equal
from glotaran.utils.ipython import MarkdownStr
from glotaran.utils.sanitize import pretty_format_numerical
from glotaran.utils.sanitize import sanitize_parameter_list

if TYPE_CHECKING:
    from glotaran.parameter import Parameters

RESERVED_LABELS: list[str] = list(asteval.make_symbol_table().keys()) + ["parameters", "iteration"]


OPTION_NAMES_SERIALIZED = {
    "expression": "expr",
    "maximum": "max",
    "minimum": "min",
    "non_negative": "non-negative",
    "standard_error": "standard-error",
}

OPTION_NAMES_DESERIALIZED = {v: k for k, v in OPTION_NAMES_SERIALIZED.items()}


def deserialize_options(options: dict[str, Any]) -> dict[str, Any]:
    """Replace options keys in serialized format by attribute names.

    Parameters
    ----------
    options : dict[str, Any]
        The serialized options.

    Returns
    -------
    dict[str, Any]
        The deserialized options.

    """
    return {OPTION_NAMES_DESERIALIZED.get(k, k): v for k, v in options.items()}


def serialize_options(options: dict[str, Any]) -> dict[str, Any]:
    """Replace options keys with serialized format by attribute names.

    Parameters
    ----------
    options : dict[str, Any]
        The options.

    Returns
    -------
    dict[str, Any]
        The serialized options.

    """
    return {OPTION_NAMES_SERIALIZED.get(k, k): v for k, v in options.items()}


PARAMETER_EXPRESSION_REGEX = re.compile(r"\$(?P<parameter_expression>[\w\d\.]+)((?![\w\d\.]+)|$)")
"""A regular expression to find and replace parameter names in expressions."""
VALID_LABEL_REGEX = re.compile(r"\W", flags=re.ASCII)
"""A regular expression to validate labels."""


def valid_label(parameter: Parameter, attribute: Attribute, label: str):
    """Check if a label is a valid label for :class:`Parameter`.

    Parameters
    ----------
    parameter : Parameter
        The :class:`Parameter` instance
    attribute : Attribute
        The label field.
    label : str
        The label value.

    Raises
    ------
    ValueError
        Raise when the label is not valid.
    """
    if VALID_LABEL_REGEX.search(label.replace(".", "_")) is not None or label in RESERVED_LABELS:
        raise ValueError(f"'{label}' is not a valid parameter label.")


def set_transformed_expression(parameter: Parameter, attribute: Attribute, expression: str | None):
    """Set the transformed expression from an expression.

    Parameters
    ----------
    parameter : Parameter
        The :class:`Parameter` instance
    attribute : Attribute
        The label field.
    expression : str | None
        The expression value.
    """
    if expression:
        parameter.vary = False
        parameter.transformed_expression = PARAMETER_EXPRESSION_REGEX.sub(
            r"parameters.get('\g<parameter_expression>').value", expression
        )


@no_default_vals_in_repr
@define
class Parameter(_SupportsArray):
    """A parameter for optimization."""

    label: str = ib(converter=str, validator=[valid_label])
    value: float = ib(
        default=np.nan,
        converter=lambda v: float(v) if isinstance(v, int) else v,
        validator=[validators.instance_of(float)],
    )
    standard_error: float = np.nan
    expression: str | None = ib(default=None, validator=[set_transformed_expression])
    maximum: float = ib(default=np.inf, validator=[validators.instance_of((int, float))])
    minimum: float = ib(default=-np.inf, validator=[validators.instance_of((int, float))])
    non_negative: bool = False
    vary: bool = ib(default=True)

    transformed_expression: str | None = ib(default=None, init=False, repr=False)

    @property
    def label_short(self) -> str:
        """Get short label.

        Returns
        -------
        str :
            The short label.
        """
        return self.label.split(".")[-1]

    @classmethod
    def from_list(
        cls,
        values: list[Any],
        *,
        default_options: dict[str, Any] | None = None,
    ) -> Parameter:
        """Create a parameter from a list.

        Parameters
        ----------
        values : list[Any]
            The list of parameter definitions.
        default_options : dict[str, Any] | None
            A dictionary of default options.

        Returns
        -------
        Parameter
            The created :class:`Parameter`.
        """
        options = None

        values = sanitize_parameter_list(values.copy())
        param = {
            "label": _retrieve_item_from_list_by_type(values, str, ""),
            "value": _retrieve_item_from_list_by_type(values, (int, float), np.nan),
        }
        options = _retrieve_item_from_list_by_type(values, dict, {})

        if default_options:
            param |= deserialize_options(default_options)
        param |= deserialize_options(options)

        return cls(**param)

    def copy(self) -> Parameter:
        """Create a copy of the :class:`Parameter`.

        Returns
        -------
        Parameter :
            A copy of the :class:`Parameter`.
        """
        return evolve(self)

    def as_dict(self) -> dict[str, Any]:
        """Get the parameter as a dictionary.

        Returns
        -------
        dict[str, Any]
            The parameter as dictionary.
        """
        return asdict(self, filter=filters.exclude(fields(Parameter).transformed_expression))

    def _deep_equals(self, other: Parameter) -> bool:
        """Compare all attributes for equality not only ``value`` like ``__eq__`` does.

        This is used by ``Parameters`` to check for equality.

        Parameters
        ----------
        other: Parameter
            Other parameter to compare against.

        Returns
        -------
        bool
            Whether or not all attributes are equal.
        """
        return all(
            nan_or_equal(self_val, other_val)
            for self_val, other_val in zip(self.as_dict().values(), other.as_dict().values())
        )

    def as_list(self, label_short: bool = False) -> list[str | float | dict[str, Any]]:
        """Get the parameter as a dictionary.

        Parameters
        ----------
        label_short : bool
            If true, the label will be replaced by the shortened label.

        Returns
        -------
        dict[str, Any]
            The parameter as dictionary.
        """
        options = self.as_dict()

        label = options.pop("label")
        value = options.pop("value")

        if label_short:
            label = self.label_short

        return [label, value, serialize_options(options)]

    def get_value_and_bounds_for_optimization(self) -> tuple[float, float, float]:
        """Get the parameter value and bounds with expression and non-negative constraints applied.

        Returns
        -------
        tuple[float, float, float]
            A tuple containing the value, the lower and the upper bound.
        """
        value = self.value
        minimum = self.minimum
        maximum = self.maximum

        if self.non_negative:
            value = _log_value(value)
            minimum = _log_value(minimum)
            maximum = _log_value(maximum)

        return value, minimum, maximum

    def set_value_from_optimization(self, value: float):
        """Set the value from an optimization result and reverses non-negative transformation.

        Parameters
        ----------
        value : float
            Value from optimization.
        """
        self.value = np.exp(value) if self.non_negative else value

    def markdown(
        self,
        all_parameters: Parameters | None = None,
        initial_parameters: Parameters | None = None,
    ) -> MarkdownStr:
        """Get a markdown representation of the parameter.

        Parameters
        ----------
        all_parameters : Parameters | None
            A parameter group containing the whole parameter set (used for expression lookup).
        initial_parameters : Parameters | None
            The initial parameter.

        Returns
        -------
        MarkdownStr
            The parameter as markdown string.
        """
        md = f"{self.label}"

        parameter = self if all_parameters is None else all_parameters.get(self.label)
        value = f"{parameter.value:.2e}"
        if parameter.vary:
            if parameter.standard_error is not np.nan:
                t_value = pretty_format_numerical(parameter.value / parameter.standard_error)
                value += f"±{parameter.standard_error:.2e}, t-value: {t_value}"

            if initial_parameters is not None:
                initial_value = initial_parameters.get(parameter.label).value
                value += f", initial: {initial_value:.2e}"
            md += f"({value})"
        elif parameter.expression is not None:
            expression = parameter.expression
            if all_parameters is not None:
                for match in PARAMETER_EXPRESSION_REGEX.findall(expression):
                    label = match[0]
                    parameter = all_parameters.get(label)
                    expression = expression.replace(
                        f"${label}", f"_{parameter.markdown(all_parameters=all_parameters)}_"
                    )

            md += f"({value}={expression})"
        else:
            md += f"({value}, fixed)"

        return MarkdownStr(md)

    def __array__(self):
        """array"""  # noqa: D400, D403
        return np.array(self.value, dtype=float)

    def __str__(self) -> str:
        """Representation used by print and str."""
        return (
            f"__{self.label}__: _Value_: {self.value}, _StdErr_: {self.standard_error}, _Min_:"
            f" {self.minimum}, _Max_: {self.maximum}, _Vary_: {self.vary},"
            f" _Non-Negative_: {self.non_negative}, _Expr_: {self.expression}"
        )

    def __abs__(self):
        """abs"""  # noqa: D400, D403
        return abs(self.value)

    def __neg__(self):
        """neg"""  # noqa: D400, D403
        return -self.value

    def __pos__(self):
        """positive"""  # noqa: D400, D403
        return +self.value

    def __int__(self):
        """int"""  # noqa: D400, D403
        return int(self.value)

    def __float__(self):
        """float"""  # noqa: D400, D403
        return float(self.value)

    def __trunc__(self):
        """trunc"""  # noqa: D400, D403
        return self.value.__trunc__()

    def __add__(self, other):
        """+"""  # noqa: D400
        return self.value + other

    def __sub__(self, other):
        """-"""  # noqa: D400
        return self.value - other

    def __truediv__(self, other):
        """/"""  # noqa: D400
        return self.value / other

    def __floordiv__(self, other):
        """//"""  # noqa: D400
        return self.value // other

    def __divmod__(self, other):
        """divmod"""  # noqa: D400, D403
        return divmod(self.value, other)

    def __mod__(self, other):
        """%"""  # noqa: D400
        return self.value % other

    def __mul__(self, other):
        """*"""  # noqa: D400
        return self.value * other

    def __pow__(self, other):
        """**"""  # noqa: D400
        return self.value**other

    def __gt__(self, other):
        """>"""  # noqa: D400
        return self.value > other

    def __ge__(self, other):
        """>="""  # noqa: D400
        return self.value >= other

    def __le__(self, other):
        """<="""  # noqa: D400
        return self.value <= other

    def __lt__(self, other):
        """<"""  # noqa: D400
        return self.value < other

    def __eq__(self, other):
        """=="""  # noqa: D400
        return self.value == other

    def __ne__(self, other):
        """!="""  # noqa: D400
        return self.value != other

    def __radd__(self, other):
        """+ (right)"""  # noqa: D400
        return other + self.value

    def __rtruediv__(self, other):
        """/ (right)"""  # noqa: D400
        return other / self.value

    def __rdivmod__(self, other):
        """divmod (right)"""  # noqa: D400, D403
        return divmod(other, self.value)

    def __rfloordiv__(self, other):
        """// (right)"""  # noqa: D400
        return other // self.value

    def __rmod__(self, other):
        """% (right)"""  # noqa: D400
        return other % self.value

    def __rmul__(self, other):
        """* (right)"""  # noqa: D400
        return other * self.value

    def __rpow__(self, other):
        """** (right)"""  # noqa: D400
        return other**self.value

    def __rsub__(self, other):
        """- (right)"""  # noqa: D400
        return other - self.value


def _log_value(value: float) -> float:
    """Get the logarithm of a value.

    Performs a check for edge cases and migitates numerical issues.

    Parameters
    ----------
    value : float
        The initial value.

    Returns
    -------
    float
        The logarithm of the value.
    """
    if not np.isfinite(value):
        return value
    if value == 1:
        value += 1e-10
    return np.log(value)


def _retrieve_item_from_list_by_type(
    item_list: list, item_type: type | tuple[type, ...], default: Any
) -> Any:
    """Retrieve an item from list which matches a given type.

    Parameters
    ----------
    item_list : list
        The list to retrieve from.
    item_type : type | tuple[type, ...]
        The item type or tuple of types to match.
    default : Any
        Returned if no item matches.

    Returns
    -------
    Any

    """
    tmp = list(filter(lambda x: isinstance(x, item_type), item_list))
    if not tmp:
        return default
    item_list.remove(tmp[0])
    return tmp[0]



from copy import deepcopy

import numpy as np
import pytest

from glotaran.model import EqualAreaPenalty
from glotaran.optimization.optimization_group import OptimizationGroup
from glotaran.optimization.test.suites import TwoCompartmentDecay as suite
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate


@pytest.mark.parametrize("index_dependent", [True, False])
@pytest.mark.parametrize("link_clp", [True, False])
def test_penalties(index_dependent, link_clp):
    model = deepcopy(suite.model)
    model.dataset_groups["default"].link_clp = link_clp
    model.megacomplex["m1"].is_index_dependent = index_dependent
    model.clp_penalties.append(
        EqualAreaPenalty(
            **{
                "source": "s1",
                "source_intervals": [(1, 20)],
                "target": "s2",
                "target_intervals": [(20, 45)],
                "parameter": "3",
                "weight": 10,
            }
        )
    )
    parameters = Parameters.from_list([11e-4, 22e-5, 2])

    global_axis = np.arange(50)

    print(f"{link_clp=}\n{index_dependent=}")  # T201
    dataset = simulate(
        suite.sim_model,
        "dataset1",
        parameters,
        {"global": global_axis, "model": suite.model_axis},
    )
    scheme = Scheme(model=model, parameters=parameters, data={"dataset1": dataset})
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])
    optimization_group.calculate(parameters)
    additional_penalty = optimization_group.get_additional_penalties()
    full_penalty = optimization_group.get_full_penalty()

    assert isinstance(additional_penalty, list)
    assert len(additional_penalty) == 1
    assert additional_penalty[0] != 0
    assert isinstance(full_penalty, np.ndarray)
    assert full_penalty.size == (suite.model_axis.size * global_axis.size) + len(
        additional_penalty
    )

    # 2 compartments * 50 items in global axis
    assert optimization_group.number_of_clps == 100



"""The glotaran data registry module."""
from __future__ import annotations

from pathlib import Path

import xarray as xr

from glotaran.io import load_dataset
from glotaran.io import save_dataset
from glotaran.plugin_system.data_io_registration import supported_file_extensions_data_io
from glotaran.project.project_registry import ProjectRegistry


class ProjectDataRegistry(ProjectRegistry):
    """A registry for data."""

    def __init__(self, directory: Path):
        """Initialize a data registry.

        Parameters
        ----------
        directory : Path
            The registry directory.
        """
        super().__init__(
            directory / "data",
            supported_file_extensions_data_io("load_dataset"),
            load_dataset,
            item_name="Dataset",
        )

    def import_data(
        self,
        dataset: str | Path | xr.Dataset | xr.DataArray,
        dataset_name: str | None = None,
        allow_overwrite: bool = False,
        ignore_existing: bool = False,
    ):
        """Import a dataset.

        Parameters
        ----------
        dataset : str | Path | xr.Dataset | xr.DataArray
            Dataset instance or path to a dataset.
        dataset_name : str | None
            The name of the dataset (needs to be provided when dataset is an xarray instance).
            Defaults to None.
        allow_overwrite: bool
            Whether to overwrite an existing dataset.
        ignore_existing: bool
            Whether to ignore import if the dataset already exists.

        Raises
        ------
        ValueError
            When importing from xarray object and not providing a name.
        """
        if isinstance(dataset, (xr.DataArray, xr.Dataset)) and dataset_name is None:
            raise ValueError(
                "When importing data from a 'xarray.Dataset' or 'xarray.DataArray' "
                "it is required to also pass a ``dataset_name``."
            )
        if isinstance(dataset, xr.DataArray):
            dataset = dataset.to_dataset(name="data")

        if isinstance(dataset, (str, Path)):
            dataset = Path(dataset)

            if dataset.is_absolute() is False:
                dataset = (self.directory.parent / dataset).resolve()

            dataset_name = dataset_name or dataset.stem
            dataset = load_dataset(dataset)

        data_path = self.directory / f"{dataset_name}.nc"
        if data_path.exists() and ignore_existing and allow_overwrite is False:
            return
        save_dataset(dataset, data_path, allow_overwrite=allow_overwrite)



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

import glotaran.builtin.io.yml.yml as yml_module
from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.io import load_model

if TYPE_CHECKING:
    from typing import Any

    from _pytest.monkeypatch import MonkeyPatch


@pytest.mark.parametrize(
    "model_yml_str, expected_nr_of_warnings, expected_key, expected_value",
    (
        (
            dedent(
                """
                clp_area_penalties:
                    - type: equal_area
                """
            ),
            1,
            "clp_penalties",
            [{"type": "equal_area"}],
        ),
    ),
    ids=("clp_area_penalties",),
)
def test_model_spec_deprecations(
    monkeypatch: MonkeyPatch,
    model_yml_str: str,
    expected_nr_of_warnings: int,
    expected_key: str,
    expected_value: Any,
):
    """Warning gets emitted by load_model"""
    return_dicts = []
    with monkeypatch.context() as m:
        m.setattr(yml_module, "sanitize_yaml", lambda spec: return_dicts.append(spec))
        record, _ = deprecation_warning_on_call_test_helper(
            load_model, args=(model_yml_str,), kwargs={"format_name": "yml_str"}
        )

        return_dict = return_dicts[0]

        assert expected_key in return_dict
        assert return_dict[expected_key] == expected_value

        assert len(record) == expected_nr_of_warnings



import numpy as np
import pytest
from IPython.core.formatters import format_display_data

from glotaran.builtin.megacomplexes.decay.k_matrix import KMatrix
from glotaran.builtin.megacomplexes.decay.k_matrix import calculate_gamma
from glotaran.model.item import fill_item
from glotaran.parameter import Parameters


class SequentialModel:
    params = [0.55, 0.0404]
    compartments = ["s1", "s2"]
    matrix = {
        ("s2", "s1"): "1",
        ("s2", "s2"): "2",
    }
    jvec = [1, 0]

    wanted_array = np.asarray(
        [
            [0, 0],
            [0.55, 0.0404],
        ]
    )

    wanted_full = np.asarray(
        [
            [-0.55, 0],
            [0.55, -0.0404],
        ]
    )

    wanted_eigen_vals = np.asarray([-0.55, -0.0404])

    wanted_eigen_vec = np.asarray(
        [
            [0.6796527, 0],
            [-0.7335341, 1],
        ]
    )

    wanted_gamma = np.diag([1.47134, 1.079278])

    wanted_a_matrix = np.asarray(
        [
            [1, -1.079278],
            [0, 1.079278],
        ]
    )


class SequentialModelWithBacktransfer:
    params = [0.55, 0.0404, 0.11]
    compartments = ["s1", "s2"]
    matrix = {
        ("s2", "s1"): "1",
        ("s1", "s2"): "3",
        ("s2", "s2"): "2",
    }
    jvec = [1, 0]

    wanted_array = np.asarray(
        [
            [0, 0.11],
            [0.55, 0.0404],
        ]
    )

    wanted_full = np.asarray(
        [
            [-0.55, 0.11],
            [0.55, -0.1504],
        ]
    )

    wanted_eigen_vals = np.asarray([-0.6670912, -0.03330879])

    wanted_eigen_vec = np.asarray(
        [
            [-0.6846927, -0.2082266],
            [0.7288318, -0.9780806],
        ]
    )

    wanted_gamma = np.diag([-1.19068, -0.8872538])

    wanted_a_matrix = np.asarray(
        [
            [0.8152501, -0.8678057],
            [0.1847499, 0.8678057],
        ]
    )


class ParallelModel:
    params = [0.55, 0.0404]
    compartments = ["s1", "s2"]
    matrix = {
        ("s1", "s1"): "1",
        ("s2", "s2"): "2",
    }
    jvec = [1, 1]

    wanted_array = np.asarray(
        [
            [0.55, 0],
            [0, 0.0404],
        ]
    )

    wanted_full = np.asarray(
        [
            [-0.55, 0],
            [0, -0.0404],
        ]
    )

    wanted_eigen_vals = np.asarray([-0.55, -0.0404])

    wanted_eigen_vec = np.asarray(
        [
            [1, 0],
            [0, 1],
        ]
    )

    wanted_gamma = np.diag([1, 1])

    wanted_a_matrix = np.asarray(
        [
            [1, 0],
            [0, 1],
        ]
    )


class ParallelModelWithEquilibria:
    params = [0.55, 0.0404, 0.11, 0.02]
    compartments = ["s1", "s2"]
    matrix = {
        ("s1", "s1"): "4",
        ("s2", "s1"): "1",
        ("s2", "s2"): "2",
        ("s1", "s2"): "3",
    }
    jvec = [1, 1]

    wanted_array = np.asarray(
        [
            [0.02, 0.11],
            [0.55, 0.0404],
        ]
    )

    wanted_full = np.asarray(
        [
            [-0.57, 0.11],
            [0.55, -0.1504],
        ]
    )

    wanted_eigen_vals = np.asarray([-0.6834894, -0.03691059])

    wanted_eigen_vec = np.asarray(
        [
            [-0.6959817, -0.2020870],
            [0.7180595, -0.9793676],
        ]
    )

    wanted_gamma = np.diag([-0.940184, -1.710398])

    wanted_a_matrix = np.asarray(
        [
            [0.6543509, -0.6751081],
            [0.3456491, 1.6751081],
        ]
    )


@pytest.mark.parametrize(
    "matrix",
    [SequentialModel, SequentialModelWithBacktransfer, ParallelModel, ParallelModelWithEquilibria],
)
def test_a_matrix_general(matrix):
    params = Parameters.from_list(matrix.params)

    mat = KMatrix(label="", matrix=matrix.matrix)
    mat = fill_item(mat, None, params)

    initial_concentration = matrix.jvec

    for comp in matrix.compartments:
        assert comp in mat.involved_compartments()

    print(mat.reduced(matrix.compartments))
    assert np.array_equal(mat.reduced(matrix.compartments), matrix.wanted_array)

    print(mat.full(matrix.compartments).T)
    assert np.allclose(mat.full(matrix.compartments), matrix.wanted_full)

    print(mat.eigen(matrix.compartments)[0])
    print(mat.eigen(matrix.compartments)[1])
    vals, vec = mat.eigen(matrix.compartments)
    assert np.allclose(vals, matrix.wanted_eigen_vals)
    assert np.allclose(vec, matrix.wanted_eigen_vec)

    print(calculate_gamma(vec, initial_concentration))
    assert np.allclose(calculate_gamma(vec, initial_concentration), matrix.wanted_gamma)

    print(mat.a_matrix_general(matrix.compartments, initial_concentration))
    assert np.allclose(
        mat.a_matrix_general(matrix.compartments, initial_concentration),
        matrix.wanted_a_matrix,
    )


def test_a_matrix_sequential():
    compartments = ["s1", "s2", "s3"]
    matrix = {
        ("s2", "s1"): "1",
        ("s3", "s2"): "2",
        ("s2", "s2"): "2",
        ("s3", "s3"): "3",
    }

    params = Parameters.from_list([3, 4, 5])
    mat = KMatrix(label="", matrix=matrix)
    mat = fill_item(mat, None, params)

    initial_concentration = [1, 0, 0]

    assert not mat.is_sequential(compartments, initial_concentration)

    matrix = {
        ("s2", "s1"): "1",
        ("s2", "s2"): "2",
    }

    compartments = ["s1", "s2"]
    params = Parameters.from_list([0.55, 0.0404])
    mat = KMatrix(label="", matrix=matrix)
    mat = fill_item(mat, None, params)

    initial_concentration = [1, 0]

    print(mat.reduced(compartments))
    assert mat.is_sequential(compartments, initial_concentration)

    wanted_a_matrix = np.asarray(
        [
            [1, -1.079278],
            [0, 1.079278],
        ]
    )

    print(mat.a_matrix_sequential(compartments))
    assert np.allclose(mat.a_matrix_sequential(compartments), wanted_a_matrix)


def test_combine_matrices():
    matrix1 = {
        ("s1", "s1"): "1",
        ("s2", "s2"): "2",
    }
    mat1 = KMatrix(label="A", matrix=matrix1)
    matrix2 = {
        ("s2", "s2"): "3",
        ("s3", "s3"): "4",
    }
    mat2 = KMatrix(label="B", matrix=matrix2)

    combined = mat1.combine(mat2)

    assert combined.label == "A+B"
    assert combined.matrix[("s1", "s1")] == "1"
    assert combined.matrix[("s2", "s2")] == "3"
    assert combined.matrix[("s3", "s3")] == "4"


def test_kmatrix_ipython_rendering():
    """Autorendering in ipython"""

    matrix = {
        ("s1", "s1"): "1",
        ("s2", "s2"): "2",
    }
    kmatrix = KMatrix(label="A", matrix=matrix)

    rendered_obj = format_display_data(kmatrix)[0]

    test_markdown_str = "text/markdown"
    assert test_markdown_str in rendered_obj
    assert rendered_obj[test_markdown_str].startswith("| compartment")

    rendered_markdown_return = format_display_data(kmatrix.matrix_as_markdown())[0]

    assert test_markdown_str in rendered_markdown_return
    assert rendered_markdown_return[test_markdown_str].startswith("| compartment")



"""Tests for glotaran/utils/io.py"""
from __future__ import annotations

import html
import os
import sys
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from IPython.core.formatters import format_display_data
from pandas.testing import assert_frame_equal

from glotaran.io import load_dataset
from glotaran.io import save_dataset
from glotaran.optimization.optimize import optimize
from glotaran.project.result import Result
from glotaran.testing.simulated_data.sequential_spectral_decay import SCHEME
from glotaran.testing.simulated_data.shared_decay import SPECTRAL_AXIS
from glotaran.utils.io import DatasetMapping
from glotaran.utils.io import chdir_context
from glotaran.utils.io import create_clp_guide_dataset
from glotaran.utils.io import load_datasets
from glotaran.utils.io import relative_posix_path
from glotaran.utils.io import safe_dataframe_fillna
from glotaran.utils.io import safe_dataframe_replace


@pytest.fixture
def ds_mapping() -> DatasetMapping:
    """Dummy mapping for testing."""
    ds_mapping = DatasetMapping()

    ds_mapping["ds1"] = xr.DataArray([1, 2]).to_dataset(name="data")
    ds_mapping["ds2"] = xr.DataArray([3, 4]).to_dataset(name="data")
    return ds_mapping


@pytest.fixture
def dummy_datasets(tmp_path: Path) -> tuple[Path, xr.Dataset, xr.Dataset]:
    """Dummy files for testing."""
    ds1 = xr.DataArray([1, 2]).to_dataset(name="data")
    ds2 = xr.DataArray([3, 4]).to_dataset(name="data")
    save_dataset(ds1, tmp_path / "ds1_file.nc")
    save_dataset(ds2, tmp_path / "ds2_file.nc")
    return tmp_path, ds1, ds2


@pytest.fixture(scope="session")
def dummy_result():
    """Dummy result for testing."""
    print(SCHEME.data["dataset_1"])
    yield optimize(SCHEME, raise_exception=True)


def test_dataset_mapping(ds_mapping: DatasetMapping):
    """Basic mapping functionality of ``DatasetMapping``."""

    assert "ds1" in ds_mapping
    assert "ds2" in ds_mapping
    assert len(ds_mapping) == 2

    assert repr(ds_mapping) == "{'ds1': <xarray.Dataset>, 'ds2': <xarray.Dataset>}"

    for ds_name, expected_ds_name in zip(ds_mapping, ["ds1", "ds2"]):
        assert ds_name == expected_ds_name

    del ds_mapping["ds1"]

    assert "ds1" not in ds_mapping
    assert "ds2" in ds_mapping
    assert len(ds_mapping) == 1

    assert repr(ds_mapping) == "{'ds2': <xarray.Dataset>}"


def test_dataset_mapping_ipython_render(ds_mapping: DatasetMapping):
    """Renders as html in an ipython context."""

    rendered_result = format_display_data(ds_mapping)[0]

    assert "text/html" in rendered_result
    assert html.unescape(rendered_result["text/html"]).startswith(
        "<pre>{'ds1': <xarray.Dataset>, 'ds2': <xarray.Dataset>}</pre>"
        "\n<details><summary>ds1</summary>"
    )
    assert rendered_result["text/plain"] == repr(ds_mapping)


def test_load_datasets_single_dataset(dummy_datasets: tuple[Path, xr.Dataset, xr.Dataset]):
    """Functionality of ``load_datasets`` with a single dataset of all supported types."""
    tmp_path, ds1, _ = dummy_datasets
    expected_source_path = (tmp_path / "ds1_file.nc").as_posix()

    str_result = load_datasets((tmp_path / "ds1_file.nc").as_posix())

    assert "ds1_file" in str_result
    assert np.all(str_result["ds1_file"].data == ds1.data)
    assert str_result["ds1_file"].source_path == expected_source_path
    assert str_result.source_path["ds1_file"] == expected_source_path

    path_result = load_datasets(tmp_path / "ds1_file.nc")

    assert "ds1_file" in path_result
    assert np.all(path_result["ds1_file"].data == ds1.data)
    assert path_result["ds1_file"].source_path == expected_source_path
    assert path_result.source_path["ds1_file"] == expected_source_path

    dataset_result = load_datasets(ds1)

    assert "ds1_file" in dataset_result
    assert np.all(dataset_result["ds1_file"].data == ds1.data)
    assert dataset_result["ds1_file"].source_path == expected_source_path
    assert dataset_result.source_path["ds1_file"] == expected_source_path

    dataarray_result = load_datasets(xr.DataArray([1, 2]))

    assert "dataset_1" in dataarray_result
    assert np.all(dataarray_result["dataset_1"].data == ds1.data)
    assert dataarray_result["dataset_1"].source_path == "dataset_1.nc"
    assert dataarray_result.source_path["dataset_1"] == "dataset_1.nc"

    pure_dataset_result = load_datasets(xr.DataArray([1, 2]).to_dataset(name="data"))

    assert "dataset_1" in pure_dataset_result
    assert np.all(pure_dataset_result["dataset_1"].data == ds1.data)
    assert pure_dataset_result["dataset_1"].source_path == "dataset_1.nc"
    assert pure_dataset_result.source_path["dataset_1"] == "dataset_1.nc"


def test_load_datasets_sequence(dummy_datasets: tuple[Path, xr.Dataset, xr.Dataset]):
    """Functionality of ``load_datasets`` with a sequence."""
    tmp_path, ds1, ds2 = dummy_datasets

    result = load_datasets([tmp_path / "ds1_file.nc", tmp_path / "ds2_file.nc"])

    assert "ds1_file" in result
    assert np.all(result["ds1_file"].data == ds1.data)
    assert result["ds1_file"].source_path == (tmp_path / "ds1_file.nc").as_posix()
    assert result.source_path["ds1_file"] == (tmp_path / "ds1_file.nc").as_posix()

    assert "ds2_file" in result
    assert np.all(result["ds2_file"].data == ds2.data)
    assert result["ds2_file"].source_path == (tmp_path / "ds2_file.nc").as_posix()
    assert result.source_path["ds2_file"] == (tmp_path / "ds2_file.nc").as_posix()


def test_load_datasets_mapping(dummy_datasets: tuple[Path, xr.Dataset, xr.Dataset]):
    """Functionality of ``load_datasets`` with a mapping."""
    tmp_path, ds1, ds2 = dummy_datasets

    result = load_datasets({"ds1": tmp_path / "ds1_file.nc", "ds2": tmp_path / "ds2_file.nc"})

    assert "ds1" in result
    assert np.all(result["ds1"].data == ds1.data)
    assert result["ds1"].source_path == (tmp_path / "ds1_file.nc").as_posix()
    assert result.source_path["ds1"] == (tmp_path / "ds1_file.nc").as_posix()

    assert "ds2" in result
    assert np.all(result["ds2"].data == ds2.data)
    assert result["ds2"].source_path == (tmp_path / "ds2_file.nc").as_posix()
    assert result.source_path["ds2"] == (tmp_path / "ds2_file.nc").as_posix()


def test_load_datasets_wrong_type():
    """Raise TypeError for not supported type"""
    with pytest.raises(
        TypeError,
        match=(
            r"Type 'int' for 'dataset_mappable' of value "
            r"'1' is not supported\."
            r"\nSupported types are:\n"
        ),
    ):
        load_datasets(1)


@pytest.mark.parametrize("rel_file_path", ("file.txt", "folder/file.txt"))
def test_relative_posix_path(tmp_path: Path, rel_file_path: str):
    """All possible permutation for the input values."""
    full_path = tmp_path / rel_file_path

    result_str = relative_posix_path(str(full_path))

    assert result_str == full_path.as_posix()

    result_path = relative_posix_path(full_path)

    assert result_path == full_path.as_posix()

    rel_result_str = relative_posix_path(str(full_path), tmp_path)

    assert rel_result_str == rel_file_path

    rel_result_path = relative_posix_path(full_path, str(tmp_path))

    assert rel_result_path == rel_file_path

    rel_result_no_common = relative_posix_path(
        (tmp_path / f"../{rel_file_path}").resolve().as_posix(), str(tmp_path)
    )

    assert rel_result_no_common == f"../{rel_file_path}"

    result_folder = tmp_path / "results"
    with chdir_context(result_folder):
        original_rel_path = relative_posix_path(rel_file_path, result_folder)
        assert original_rel_path == rel_file_path

        original_rel_path = relative_posix_path(rel_file_path, "not_a_parent")
        assert original_rel_path == rel_file_path


def test_chdir_context(tmp_path: Path):
    """Original Path is restored even after exception is thrown."""
    original_dir = Path(os.curdir).resolve()
    with chdir_context(tmp_path) as curdir:
        assert curdir == tmp_path.resolve()
        assert tmp_path.resolve() == Path(os.curdir).resolve()
        assert Path("test.txt").resolve() == (tmp_path / "test.txt").resolve()

    assert Path(os.curdir).resolve() == original_dir

    with pytest.raises(ValueError):
        with chdir_context(tmp_path):
            raise ValueError("Original path will be restored after I raise.")

    assert Path(os.curdir).resolve() == original_dir


def test_chdir_context_exception(tmp_path: Path):
    """Raise error if ``folder_path`` is an existing file instead of a folder."""
    file_path = tmp_path / "test.txt"
    file_path.touch()

    with pytest.raises(ValueError) as excinfo:
        with chdir_context(file_path):
            pass

    assert (
        str(excinfo.value)
        == "Value of 'folder_path' needs to be a folder but was an existing file."
    )


@pytest.mark.skipif(not sys.platform.startswith("win32"), reason="Only needed for Windows")
def test_relative_posix_path_windows_diff_drives():
    """os.path.relpath doesn't cause crash when files are on different drives."""

    source_path = "D:\\data\\data_file.txt"
    result = relative_posix_path(source_path, "C:\\result_path")

    assert result == Path(source_path).as_posix()


def test_safe_dataframe_fillna():
    """Only values in selected columns are filled"""
    df = pd.DataFrame(
        {
            "stays_same": [np.nan, -4, 2],
            "minimum": [np.nan, -4, 2],
            "maximum": [np.nan, -4, 2],
        }
    )

    df2 = pd.DataFrame(
        {
            "stays_same": [np.nan, -4, 2],
            "minimum": [-np.inf, -4, 2],
            "maximum": [np.inf, -4, 2],
        }
    )

    safe_dataframe_fillna(df, "minimum", -np.inf)
    safe_dataframe_fillna(df, "maximum", np.inf)
    safe_dataframe_fillna(df, "not_a_column", np.inf)

    assert_frame_equal(df, df2)


def test_safe_dataframe_replace():
    """Only values in selected columns are replaced"""

    df = pd.DataFrame(
        {
            "stays_same": [np.nan, -4, 2],
            "minimum": [-np.inf, -4, 2],
            "maximum": [np.inf, -4, 2],
            "list_test": [np.inf, -4, 2],
            "tuple_test": [np.inf, -4, 2],
        }
    )

    df2 = pd.DataFrame(
        {
            "stays_same": [np.nan, -4, 2],
            "minimum": ["", -4, 2],
            "maximum": ["", -4, 2],
            "list_test": [1.0, 1, 2],
            "tuple_test": [3.0, 3, 2],
        }
    )

    safe_dataframe_replace(df, "minimum", -np.inf, "")
    safe_dataframe_replace(df, "maximum", (np.inf, np.nan), "")
    safe_dataframe_replace(df, "list_test", (np.inf, -4), 1)
    safe_dataframe_replace(df, "tuple_test", (np.inf, -4), 3)
    safe_dataframe_replace(df, "not_a_column", np.inf, 2)

    assert_frame_equal(df, df2)


def test_create_clp_guide_dataset(dummy_result: Result):
    """Check that clp guide has correct dimensions and dimension values."""
    clp_guide = create_clp_guide_dataset(dummy_result, "species_1", "dataset_1")

    assert clp_guide.data.shape == (1, dummy_result.data["dataset_1"].spectral.size)
    assert np.allclose(clp_guide.coords["time"].item(), -1)
    assert np.allclose(clp_guide.coords["spectral"].values, SPECTRAL_AXIS)

    clp_guide = create_clp_guide_dataset(dummy_result.data["dataset_1"], "species_1")

    assert clp_guide.data.shape == (1, dummy_result.data["dataset_1"].spectral.size)
    assert np.allclose(clp_guide.coords["time"].item(), -1)
    assert np.allclose(clp_guide.coords["spectral"].values, SPECTRAL_AXIS)


def test_create_clp_guide_dataset_errors(dummy_result: Result):
    """Errors thrown when dataset or clp_label are not in result."""
    with pytest.raises(ValueError) as exc_info:
        create_clp_guide_dataset(dummy_result, "species_1", "not-a-dataset")

    assert (
        str(exc_info.value)
        == "Unknown dataset 'not-a-dataset'. Known datasets are:\n ['dataset_1']"
    )

    with pytest.raises(ValueError) as exc_info:
        create_clp_guide_dataset(dummy_result, "not-a-species", "dataset_1")

    assert (
        str(exc_info.value) == "Unknown clp_label 'not-a-species'. Known clp_labels are:\n "
        "['species_1', 'species_2', 'species_3']"
    )

    dummy_dataset = dummy_result.data["dataset_1"].copy()
    del dummy_dataset.attrs["model_dimension"]

    with pytest.raises(ValueError) as exc_info:
        create_clp_guide_dataset(dummy_dataset, "species_1")

    assert (
        str(exc_info.value) == "Result dataset is missing attribute 'model_dimension', "
        "which means that it was created with pyglotaran<0.6.0."
        "Please recreate the result with the latest version of pyglotaran."
    )


def test_extract_sas_ascii_round_trip(dummy_result: Result, tmp_path: Path):
    """Save to and then load from ascii results in the same data (spectrum)."""
    tmp_file = tmp_path / "sas.ascii"

    sas = create_clp_guide_dataset(dummy_result, "species_1", "dataset_1")
    with pytest.warns(UserWarning) as rec_warn:
        save_dataset(sas, tmp_file)

        assert len(rec_warn) == 1
        assert Path(rec_warn[0].filename).samefile(__file__)
        assert rec_warn[0].message.args[0] == (
            "Saving the 'data' attribute of 'dataset' as a fallback."
            "Result saving for ascii format only supports xarray.DataArray format, "
            "please pass a xarray.DataArray instead of a xarray.Dataset (e.g. dataset.data)."
        )

    loaded_sas = load_dataset(tmp_file, prepare=False)

    for dim in sas.dims:
        assert all(sas.coords[dim] == loaded_sas.coords[dim]), f"Coordinate {dim} mismatch"
    assert np.allclose(sas.data.values, loaded_sas.data.values)
